import re
from math import cos, radians, sqrt, sin
from typing import List, Union, TYPE_CHECKING, Optional, Iterator, Tuple, Dict, Generator

from shelxfile.atoms.pairs import AtomPair
from shelxfile.misc.dsrmath import my_isnumeric, SymmetryElement, OrthogonalMatrix, Matrix
from shelxfile.misc.misc import chunks, ParseParamError, ParseNumError, \
    ParseOrderError, ParseSyntaxError

if TYPE_CHECKING:
    from shelxfile import Shelxfile
    from shelxfile.atoms.atom import Atom
"""
SHELXL cards:

ABIN n1 n2
ACTA 2θfull[#]
AFIX mn d[#] sof[11] U[10.08]
ANIS n
ANIS names
ANSC six coefficients
ANSR anres[0.001]
BASF scale factors
BIND atom1 atom2
BIND m n
BLOC n1 n2 atomnames
BOND atomnames
BUMP s [0.02]
CELL λ a b c α β γ
CGLS nls[0] nrf[0] nextra[0]
CHIV V[0] s[0.1] atomnames
CONF atomnames max_d[1.9] max_a[170]
CONN bmax[12] r[#] atomnames or CONN bmax[12]
DAMP damp[0.7] limse[15]
DANG d s[0.04] atom pairs
DEFS sd[0.02] sf[0.1] su[0.01] ss[0.04] maxsof[1]
DELU s1[0.01] s2[0.01] atomnames
DFIX d s[0.02] atom pairs
DISP E f' f"[#] mu[#]
EADP atomnames
END
EQIV $n symmetry operation
EXTI x[0]
EXYZ atomnames
FEND
FLAT s[0.1] four or more atoms
FMAP code[2] axis[#] nl[53]
FRAG code[17] a[1] b[1] c[1] α[90] β[90] γ[90]
FREE atom1 atom2
FVAR osf[1] free variables
GRID sl[#] sa[#] sd[#] dl[#] da[#] dd[#]
HFIX mn U[#] d[#] atomnames
HKLF N[0] S[1] r11...r33[1 0 0 0 1 0 0 0 1] sm[1] m[0]
HTAB dh[2.0]
HTAB donor-atom acceptor-atom
ISOR s[0.1] st[0.2] atomnames
LATT N[1]
LAUE E
LIST m[#] mult[1]
L.S. nls[0] nrf[0] nextra[0]
MERG n[2]
MORE m[1]
MOVE dx[0] dy[0] dz[0] sign[1]
MPLA na atomnames
NCSY DN sd[0.1] su[0.05] atoms
NEUT
OMIT atomnames
OMIT s[-2] 2θ(lim)[180]
OMIT h k l
PART n sof
PLAN npeaks[20] d1[#] d2[#]
PRIG p[#]
REM
RESI class[ ] number[0] alias
RIGU s1[0.004] s2[0.004] atomnames
RTAB codename atomnames
SADI s[0.02] pairs of atoms
SAME s1[0.02] s2[0.04] atomnames
SFAC elements
SFAC E a1 b1 a2 b2 a3 b3 a4 b4 c f' f" mu r wt
SHEL lowres[infinite] highres[0]
SIMU s[0.04] st[0.08] dmax[2.0] atomnames
SIZE dx dy dz
SPEC del[0.2]
STIR sres step[0.01]
SUMP c sigma c1 m1 c2 m2 ...
SWAT g[0] U[2]
SYMM symmetry operation
TEMP T[20]
TITL [ ]
TWIN 3x3 matrix [-1 0 0 0 -1 0 0 0 -1] N[2]
TWST N[0]
UNIT n1 n2 ...
WGHT a[0.1] b[0] c[0] d[0] e[0] f[.33333]
WIGL del[0.2] dU[0.2]
WPDB n[1]
XNPD Umin[-0.001]
ZERR Z esd(a) esd(b) esd(c) esd(α) esd(β) esd(γ)
"""


class Residue():
    def __init__(self):
        self.shx: Optional['Shelxfile'] = None
        self._spline: str = ''
        self.residue_class: str = ''  # '' is the default class (with residue number 0)

    @property
    def residue_number(self) -> List[int]:
        if '_' in self._spline[0] and '$' not in self._spline[0]:
            _, suffix = self._spline[0].upper().split('_')
            if suffix.isdigit():
                # TODO: implement _+, _- and _*
                if '*' in suffix:
                    return list(self.shx.residues.residue_numbers.keys())  # type: ignore
                else:
                    return [int(suffix)]
        return self.shx.residues.residue_classes.get(self.residue_class, [0])  # type: ignore


class Restraint(Residue):

    def __init__(self, shx: 'Shelxfile', spline: list):
        """
        Base class for parsing restraints.
        TODO: resolve ranges like SADI_CCF3 O1 > F9
        """
        super().__init__()
        self.shx: 'Shelxfile' = shx
        self.textline: str = ' '.join(spline)
        self.name: Optional[str] = None
        self.atoms: List[Atom] = []

    @property
    def index(self) -> int:
        return self.shx.index_of(self)

    def _parse_line(self, spline: List[str]):
        """
        Residues may be referenced by any instruction that allows atom names; the reference takes
        the form of the character '_' followed by either the residue class or number without intervening
        spaces.
        Individual atom names in an instruction may be followed by '_' and a residue number, but not by '_* ' or '_'
        and a residue class. If an atom name is not followed by a residue number, the current residue is
        assumed (unless overridden by a global residue number or class appended to the instruction
        codeword).
        """
        self._spline = spline
        if '_' in spline[0]:
            self.name, suffix = spline[0].upper().split('_')
            # a residue class has to start with a letter, but can contain numbers:
            if len(suffix) > 0 and re.match(r'[a-zA-Z]', suffix):
                self.residue_class: str = suffix
        else:
            self.name = spline[0].upper()
        self._set_defs_values()
        self._check_if_class_name_fits_to_command()
        params = []
        atoms = []
        for x in spline[1:]:
            if my_isnumeric(x):
                params.append(float(x))
            else:
                atoms.append(x)
        # if pairs:
        #    return params, self.get_atompairs(atoms)
        # else:
        return params, atoms

    def _get_atompairs(self, atoms: List[str]) -> List[AtomPair]:
        pairs = []
        for p in chunks(atoms, 2):
            pairs.append(AtomPair(*p))
        return pairs

    def _check_if_class_name_fits_to_command(self) -> None:
        if self.shx.debug and self.__class__.__name__ != self.name:
            print('*** Trying to parse restraint with wrong class ***')
            raise ParseSyntaxError(debug=self.shx.debug, verbose=self.shx.verbose)

    def _set_defs_values(self) -> None:
        # Beware! DEFS changes only the non-defined default values:
        # DEFS sd[0.02] sf[0.1] su[0.01] ss[0.04] maxsof[1]
        if self.shx.defs:  # and self.shx.defs.active:
            if self.name == 'DFIX':
                self.s = self.shx.defs.sd
            if self.name == 'SAME':
                self.s1 = self.shx.defs.sd
                self.s2 = self.shx.defs.sd * 2
            if self.name == 'SADI':
                self.s = self.shx.defs.sd
            if self.name == 'CHIV':
                self.s = self.shx.defs.sf
            if self.name == 'FLAT':
                self.s = self.shx.defs.sf
            if self.name == 'DELU':
                self.s1 = self.shx.defs.su
                self.s2 = self.shx.defs.su
            if self.name == 'SIMU':
                self.s = self.shx.defs.ss
                self.st = self.shx.defs.ss * 2

    def _paircheck(self):
        if not self.atoms:
            return
        if len(self.atoms) % 2 != 0 and (self.shx.debug or self.shx.verbose):
            print('*** Wrong number of numerical parameters ***')
            print('Instruction: {}'.format(self.textline))
            if self.shx.debug:
                raise ParseNumError(debug=self.shx.debug, verbose=self.shx.verbose)

    def __iter__(self) -> Iterator[str]:
        for x in self.textline.split():
            yield x

    def __repr__(self) -> str:
        return self.textline

    def __str__(self) -> str:
        return self.textline
        '''# print(self.atoms, self.residue_number)
        s = self.s if hasattr(self, 's') else ''
        s1 = self.s1 if hasattr(self, 's1') else ''
        s2 = self.s2 if hasattr(self, 's2') else ''
        st = self.st if hasattr(self, 'st') else ''
        return f"{self.name}" \
               f"{' ' if s else ''}{s}" \
               f"{' ' if s1 else ''}{s1}" \
               f"{' ' if s2 else ''}{s2}" \
               f"{' ' if st else ''}{st} " \
               f"{' '.join(self.atoms)}"'''

    def split(self):
        return self.textline.split()


class Command():
    """
    A class to parse all general commands except restraints.
    """

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        self._shx: 'Shelxfile' = shx
        self._spline: List[str] = spline
        self.residue_class: str = ''
        self._textline: str = ' '.join(spline)

    def _parse_line(self, spline: List[str], intnums: bool = False) -> Tuple[List[Union[int, float]], List[str]]:
        """
        :param spline: Split shelxl line
        :param intnums: if numerical parameters should be integer
        :return: Tuple of (numerical parameter, words)
        """
        if '_' in spline[0]:
            self._card_name, suffix = spline[0].upper().split('_')
        else:
            self._card_name = spline[0].upper()
        numparams = []
        words = []
        for x in spline[1:]:  # all values after SHELX card
            if str.isdigit(x[0]) or x[0] in '+-':
                if intnums:
                    numparams.append(int(x))
                else:
                    numparams.append(float(x))
            else:
                words.append(x)
        return numparams, words

    def set(self, value):
        self.__init__(self._shx, value.split())

    @property
    def index(self):
        return self._shx.index_of(self)

    @property
    def position(self) -> int:
        return self.index

    def __iter__(self):
        for x in self.__repr__().split():
            yield x

    def split(self):
        return self._textline.split()

    def __str__(self):
        return self._textline

    def __repr__(self):
        return self._textline


class ABIN(Command):

    def __init__(self, shx, spline):
        """
        ABIN n1 n2
        """
        super(ABIN, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        if len(p) > 0:
            self.n1 = p[0]
        if len(p) > 1:
            self.n2 = p[1]


class ANIS(Command):

    def __init__(self, shx, spline: List):
        """
        ANIS
        ANIS n
        ANIS names

        Make either all atoms anisotropic (with just ANIS) or specific atoms with ANIS names.
        Also, wildcards like 'ANIS $CL' would make all chlorine atoms anisotropic.
        Since ANIS, like other instructions, applies to the current residue unless otherwise specified,
        ANIS_* $S would be required to make the sulfur atoms in all residues anisotropic.
        ANIS n makes the next n isotropic non-hydrogen atoms anisotropic.
        """
        super(ANIS, self).__init__(shx, spline)
        p, self.atoms = self._parse_line(spline)
        self.over_all = True
        if len(p) > 0:
            self.over_all = False
            self.n = p[0]
        if len(self.atoms) > 0:
            self.over_all = False

    def __bool__(self):
        return True


class MPLA(Command):

    def __init__(self, shx, spline: List):
        """
        MPLA na atomnames
        """
        super(MPLA, self).__init__(shx, spline)
        p, self.atoms = self._parse_line(spline, intnums=True)
        if len(p) > 0:
            self.na = p[0]


class MORE(Command):

    def __init__(self, shx, spline: List):
        """
        MORE m[1]
        """
        super(MORE, self).__init__(shx, spline)
        self.m = 1
        p, _ = self._parse_line(spline, intnums=True)
        self.m = p[0]


class CELL(Command):

    def __init__(self, shx, spline: List):
        """
        CELL λ a b c α β γ
        """
        super(CELL, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self._cell_list = []
        if len(p) > 0:
            self.wavelen = float(p[0])
        if len(p) > 6:
            self._cell_list = p[1:]
            self.a = p[1]
            self.b = p[2]
            self.c = p[3]
            self.alpha = p[4]
            self.beta = p[5]
            self.gamma = p[6]
            self.cosal = cos(radians(self.alpha))
            self.cosbe = cos(radians(self.beta))
            self.cosga = cos(radians(self.gamma))
            self.V = self.volume
            self.o = OrthogonalMatrix(self.a, self.b, self.c, self.alpha, self.beta, self.gamma)
            # calculate reciprocal lattice vectors:
            self.astar = (self.b * self.c * sin(radians(self.alpha))) / self.V
            self.bstar = (self.a * self.c * sin(radians(self.beta))) / self.V
            self.cstar = (self.a * self.b * sin(radians(self.gamma))) / self.V
            # matrix with the reciprocal lattice vectors:
            self.N = Matrix([[self.astar, 0, 0],
                             [0, self.bstar, 0],
                             [0, 0, self.cstar]])
        else:
            raise ParseSyntaxError(debug=shx.debug, verbose=shx.verbose)

    @property
    def volume(self) -> float:
        """
        calculates the volume of a unit cell
        """
        try:
            ca, cb, cg = cos(radians(self.alpha)), cos(radians(self.beta)), cos(radians(self.gamma))
            v = self.a * self.b * self.c * sqrt(1 + 2 * ca * cb * cg - ca ** 2 - cb ** 2 - cg ** 2)
        except AttributeError:
            # No valid celll
            v = 0.0
        return v

    def __iter__(self):
        return iter(self._cell_list)

    def __getitem__(self, item: Union[slice, int]) -> Union[List[float], float]:
        return self._cell_list[item]


class ZERR(Command):

    def __init__(self, shx, spline: List):
        """
        ZERR Z esd(a) esd(b) esd(c) esd(α) esd(β) esd(γ)
        """
        super(ZERR, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.Z = 1
        if len(p) > 0:
            self.Z = p[0]
        if len(p) > 6:
            self.esd_list = p[1:]
            self.esd_a = p[0]
            self.esd_b = p[1]
            self.esd_c = p[2]
            self.esd_al = p[3]
            self.esd_be = p[4]
            self.esd_ga = p[5]


class AFIX(Command):

    def __init__(self, shx, spline: list):
        """
        AFIX mn d[#] sof[11] U[10.08]
        """
        super(AFIX, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.U = 10.08
        self.sof = 11.0
        self.mn = None
        self.d = None
        if len(p) > 0:
            self.mn = int(p[0])
        if len(p) > 1:
            self.d = p[1]
        if len(p) > 2:
            self.sof = p[2]
        if len(p) > 3:
            self.U = p[3]

    def __bool__(self):
        if self.mn and self.mn > 0:
            return True
        else:
            return False


class Residues():

    def __init__(self, shx):
        self.shx = shx
        self.all_residues: list = []
        self.residue_classes: dict = {}  # class: numbers

    def append(self, resi: 'RESI') -> None:
        """
        Adds a new residues to the list of residues.
        """
        self.all_residues.append(resi)
        # Collect dict with class: numbers
        if resi.residue_class in self.residue_classes:
            self.residue_classes[resi.residue_class].append(resi.residue_number)
        else:
            self.residue_classes[resi.residue_class] = [resi.residue_number]

    @property
    def residue_numbers(self):
        return dict((x.residue_number, x.residue_class) for x in self.shx.residues.all_residues)


class RESI(Command):

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        """
        RESI class[ ] number[0] alias
        """
        super().__init__(shx, spline)
        self.shx = shx
        self.residue_class = ''
        self.residue_number: int = 0
        self.alias: Optional[int] = None
        self.chain_id: Optional[int] = None
        self._textline: str = ' '.join(spline)
        if len(spline) < 2 and (self.shx.debug or self.shx.verbose):
            print('*** Wrong RESI definition found! Check your RESI instructions ***')
            raise ParseParamError(debug=self.shx.debug, verbose=self.shx.verbose)
        self._get_resi_definition(spline)
        if self.residue_number < -999 or self.residue_number > 9999:
            if self.shx.debug or self.shx.verbose:
                print('*** Invalid residue number given. ****')
            if self.shx.debug:
                raise ParseSyntaxError(debug=self.shx.debug, verbose=self.shx.verbose)

    def _get_resi_definition(self, resi: List[str]) -> Tuple[str, int, str, int]:
        """
        RESI class[ ] number[0] alias

        Returns the residue number and class of a string like 'RESI TOL 1'
        or 'RESI 1 TOL'

        Residue names may now begin with a digit.
        They must however contain at least one letter

        Allowed residue numbers is now from -999 to 9999 (2017/1)
        """
        alpha = re.compile('[a-zA-Z]')
        for x in resi:
            if alpha.search(x):
                if ':' in x:
                    # contains ":" thus must be a chain-id+number
                    self.chain_id, self.residue_number = x.split(':')[0], int(x.split(':')[1])
                else:
                    # contains letters, must be a name (class)
                    self.residue_class = x
            else:
                # everything else can only be a number
                if self.residue_number > 0:
                    self.alias = int(x)
                else:
                    try:
                        self.residue_number = int(x)
                    except ValueError:
                        self.residue_number = 0
        return self.residue_class, self.residue_number, self.chain_id, self.alias

    def __bool__(self) -> bool:
        if self.residue_number > 0:
            return True
        else:
            return False


class PART(Command):

    def __init__(self, shx, spline: list):
        """
        PART n sof
        """
        super(PART, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.sof = 11.0
        self.n = 0
        try:
            self.n = int(p[0])
        except(ValueError, IndexError):
            if self.shx.debug or self.shx.verbose:
                print('*** Wrong PART definition in line {} found! '
                      'Check your PART instructions ***'.format(shx.error_line_num))
                if self.shx.debug:
                    raise ParseSyntaxError(debug=True)
            self.n = 0
        if len(p) > 1:
            self.sof = float(p[1])

    def __bool__(self):
        if self.n > 0:
            return True
        else:
            return False


class XNPD(Command):

    def __init__(self, shx, spline: list):
        """
        XNPD Umin[-0.001]
        """
        super(XNPD, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.Umin = -0.001
        if len(p) > 0:
            self.Umin = p[0]


class SIZE(Command):

    def __init__(self, shx, spline: list):
        """
        SIZE dx dy dz
        """
        super(SIZE, self).__init__(shx, spline)
        self.dx, self.dy, self.dz = None, None, None
        p, _ = self._parse_line(spline)
        if len(p) > 0:
            self.dx = p[0]
        if len(p) > 1:
            self.dy = p[1]
        if len(p) > 2:
            self.dz = p[2]

    def _as_text(self):
        if all([self.dx, self.dy, self.dz]):
            return "SIZE {:,g} {:,g} {:,g}".format(self.dx, self.dy, self.dz)
        else:
            return ""

    @property
    def max(self):
        return max(self.dx, self.dy, self.dz)

    @property
    def mid(self):
        return sorted([self.dx, self.dy, self.dz])[1]

    @property
    def min(self):
        return min(self.dx, self.dy, self.dz)

    def __repr__(self):
        return self._as_text()

    def __str__(self):
        return self._as_text()


class SHEL(Command):

    def __init__(self, shx, spline: list):
        """
        SHEL lowres[infinite] highres[0]
        """
        super(SHEL, self).__init__(shx, spline)
        params, _ = self._parse_line(spline)
        self.lowres = None
        self.highres = None
        if len(params) > 0:
            self.lowres = params[0]
        if len(params) > 1:
            self.highres = params[1]


class WIGL(Command):

    def __init__(self, shx, spline: list):
        """
        WIGL del[0.2] dU[0.2]
        """
        super(WIGL, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.d = 0.2
        self.dU = 0.2
        if len(p) > 0:
            self.d = p[0]
        if len(p) > 1:
            self.dU = p[1]


class WPDB(Command):

    def __init__(self, shx, spline: list):
        """
        WPDB n[1]
        """
        super(WPDB, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.n = 1
        if len(p) > 0:
            self.n = p[0]


class SPEC(Command):

    def __init__(self, shx, spline: list):
        """
        SPEC d[0.2]
        """
        super(SPEC, self).__init__(shx, spline)
        self.d = None
        p, _ = self._parse_line(spline)
        if len(p) > 0:
            self.d = p[0]


class STIR(Command):

    def __init__(self, shx, spline: list):
        """
        STIR sres step[0.01]
        """
        super(STIR, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.step = 0.01
        self.sres = None
        if len(p) > 0:
            self.sres = p[0]
        if len(p) > 1:
            self.step = p[1]

    def __repr__(self):
        return "STIR {} {}".format(self.sres if self.sres else '', self.step)

    def __str__(self):
        return "STIR {} {}".format(self.sres if self.sres else '', self.step)


class TWST(Command):

    def __init__(self, shx, spline: list):
        """
        TWST N[0] (N[1] after SHELXL-2018/3)
        Twin component number to be used for the completeness and Friedel completeness statistics.
        """
        super(TWST, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.N = 1
        if len(p) > 0:
            self.N = p[0]


class RTAB(Command):

    def __init__(self, shx, spline: list):
        """
        RTAB codename atomnames
        """
        super(RTAB, self).__init__(shx, spline)
        self.code = spline.pop(1)
        _, self.atoms = self._parse_line(spline)


class PRIG(Command):

    def __init__(self, shx, spline: list):
        """
        PRIG p[#]
        """
        super(PRIG, self).__init__(shx, spline)
        params, _ = self._parse_line(spline)
        if len(params) > 0:
            self.p = params[0]


class PLAN(Command):

    def __init__(self, shx, spline: list):
        """
        PLAN npeaks[20] d1[#] d2[#]
        """
        self.shx = shx
        super(PLAN, self).__init__(shx, spline)
        self.npeaks = 20
        self.d1 = None
        self.d2 = None
        params, _ = self._parse_line(spline)
        if len(params) > 0:
            self.npeaks = int(params[0])
        if len(params) > 1:
            self.d1 = params[1]
        if len(params) > 2:
            self.d2 = params[2]

    def __repr__(self):
        return f'PLAN {self.npeaks:,g}{" " if self.d1 else ""}{self.d1 if self.d1 is not None else ""}' \
               f'{" " if self.d2 else ""}{self.d2 if self.d2 is not None else ""}'


class FRAG(Command):

    def __init__(self, shx, spline: list):
        """
        FRAG code[17] a[1] b[1] c[1] α[90] β[90] γ[90]
        """
        super(FRAG, self).__init__(shx, spline)
        params, _ = self._parse_line(spline)
        self.code = params[0]
        self.cell = params[1:7]


class FREE(Command):

    def __init__(self, shx, spline: list):
        """
        FREE atom1 atom2
        """
        super(FREE, self).__init__(shx, spline)
        _, atoms = self._parse_line(spline)
        self.atom1 = None
        self.atom2 = None
        try:
            self.atom1 = atoms[0]
            self.atom2 = atoms[1]
        except IndexError:
            raise ParseParamError(debug=shx.debug, verbose=shx.verbose)


class FMAP(Command):
    """
    FMAP code[2] axis[#] nl[53]
    """

    def __init__(self, shx, spline: list):
        super(FMAP, self).__init__(shx, spline)
        params, _ = self._parse_line(spline)
        self.code = None
        self.axis = None
        self.nl = None
        if len(params) > 0:
            self.code = params[0]
        if len(params) > 1:
            self.axis = params[1]
        if len(params) > 2:
            self.nl = params[2]


class MOVE(Command):

    def __init__(self, shx, spline: list):
        """
        MOVE dx[0] dy[0] dz[0] sign[1]
        """
        super(MOVE, self).__init__(shx, spline)
        params, _ = self._parse_line(spline)
        self.dxdydz = None
        self.sign = None
        if len(params) > 2:
            self.dxdydz = params[:3]
        if len(params) > 3:
            self.sign = params[3]


class MERG(Command):

    def __init__(self, shx, spline: list):
        """
        MERG n[2]
        """
        super(MERG, self).__init__(shx, spline)
        self.n = None
        _n, _ = self._parse_line(spline)
        if len(_n) > 0:
            self.n = _n[0]


class HTAB(Command):

    def __init__(self, shx, spline: list):
        """
        HTAB dh[2.0]
        HTAB donor-atom acceptor-atom
        """
        super(HTAB, self).__init__(shx, spline)
        self.dh = None
        self.donor = None
        self.acceptor = None
        dh, atoms = self._parse_line(spline)
        if dh:
            self.dh = dh[0]
        if len(atoms) == 2:
            self.donor = atoms[0]
            self.acceptor = atoms[1]


class GRID(Command):

    def __init__(self, shx, spline: list):
        """
        GRID sl[#] sa[#] sd[#] dl[#] da[#] dd[#]
        """
        super(GRID, self).__init__(shx, spline)
        params, _ = self._parse_line(spline)
        if len(params) > 0:
            self.sl = params[0]
        if len(params) > 1:
            self.sa = params[1]
        if len(params) > 2:
            self.sd = params[2]
        if len(params) > 3:
            self.dl = params[3]
        if len(params) > 4:
            self.da = params[4]
        if len(params) > 5:
            self.dd = params[5]


class ACTA(Command):

    def __init__(self, shx, spline: list):
        """
        ACTA 2θfull[#] (NOHKL)
        """
        super(ACTA, self).__init__(shx, spline)
        self.twotheta, self.nohkl = self._parse_line(spline)
        self.shx = shx

    def _as_str(self):
        if self.twotheta:
            return f"ACTA {self.twotheta[0]:,g}"
        else:
            return "ACTA"

    def __repr__(self):
        return self._as_str()

    def __str__(self):
        return self._as_str()


class BLOC(Command):

    def __init__(self, shx, spline: list):
        """
        BLOC n1 n2 atomnames
        """
        super(BLOC, self).__init__(shx, spline)
        params, self.atoms = self._parse_line(spline)
        self.n1 = None
        self.n2 = None
        if len(params) > 0:
            self.n1 = params[0]
        if len(params) > 1:
            self.n2 = params[1]
        self.shx = shx


class FVAR():
    def __init__(self, number: int = 1, value: float = 0.0):
        """
        FVAR osf[1] free variables
        """
        self.fvar_value = value  # value
        self.number = number  # occurence inside of FVAR instructions
        self.usage = 1

    def __str__(self):
        return str(float(self.fvar_value))

    def __repr__(self):
        return str(float(self.fvar_value))


class FVARs():
    def __init__(self, shx):
        super(FVARs, self).__init__()
        self.fvars = []  # free variables
        self.shx = shx
        self._fvarline = 0

    def __iter__(self):
        """
        Must be defined for __repr__() to work.
        """
        for x in self.fvars:
            yield x

    def __getitem__(self, item: int) -> str:
        # SHELXL counts fvars from 1 to x:
        item = abs(item) - 1
        return self.fvars[item].fvar_value

    def __setitem__(self, key, fvar_value):
        self.fvars[key] = fvar_value

    def __len__(self) -> int:
        return len(self.fvars)

    def __str__(self) -> str:
        # returnes FVAR as list of FVAR instructions with seven numbers in one line
        lines = chunks(self.as_stringlist, 7)
        fvars = ['   '.join(i) for i in lines]
        fvars = ['FVAR   ' + i for i in fvars]
        return "\n".join(fvars)

    def __repr__(self):
        return str([x for x in self.fvars])

    @property
    def position(self) -> int:
        return self.shx.index_of(self)

    def set_free_variables(self, fvar: int, dummy_fvar: float = 0.5):
        """
        Inserts additional free variables according to the fvar number.
        """
        if fvar > 99:
            print('*** SHELXL allows only 99 free variables! ***')
            raise ParseParamError(debug=self.shx.debug, verbose=self.shx.verbose)
        varlen = len(self.fvars)
        difference = (abs(fvar) - varlen)
        if difference > 0:
            for n in range(int(difference)):
                fv = FVAR(varlen + n, dummy_fvar)
                self.fvars.append(fv)

    def append(self, fvar) -> None:
        self.fvars.append(fvar)

    def set_fvar_usage(self, fvarnum: int, times: int = 1) -> None:
        """
        Incerements the usage count of a free variable by times.
        """
        fvarnum = abs(fvarnum)
        if len(self.fvars) >= abs(fvarnum):
            self.fvars[fvarnum - 1].usage += times
        elif fvarnum > 1 and (self.shx.debug or self.shx.verbose):
            print('*** Free variable {} is not defined but used! ***'.format(fvarnum))
            if self.shx.debug:
                raise ParseParamError(debug=self.shx.debug, verbose=self.shx.verbose)

    def get_fvar_usage(self, fvarnum):
        """
        Returns the usage (count) of a certain free variable.
        """
        try:
            usage = self.fvars[fvarnum - 1].usage
        except IndexError:
            return 0
        return usage

    def fvars_used(self):
        """
        Retruns a dictionary with the usage of all free variables.
        """
        used = {}
        for num, fv in enumerate(self.fvars):
            used[num + 1] = fv.usage
        return used

    @property
    def as_stringlist(self):
        return [str(x.fvar_value) for x in self.fvars]


class CONF(Command):

    def __init__(self, shx, spline: list) -> None:
        """
        CONF atomnames max_d[1.9] max_a[170]
        """
        super(CONF, self).__init__(shx, spline)


class CONN(Command):

    def __init__(self, shx, spline: list) -> None:
        """
        CONN bmax[12] r[#] atomnames or CONN bmax[12]
        """
        super(CONN, self).__init__(shx, spline)


class REM(Command):

    def __init__(self, shx, spline: list) -> None:
        """
        Parses REM lines
        """
        super(REM, self).__init__(shx, spline)


class BIND(Command):

    def __init__(self, shx, spline: list) -> None:
        """
        BIND atom1 atom2
        BIND m n
        """
        super(BIND, self).__init__(shx, spline)
        self.parts, self.atoms = self._parse_line(spline)


class BOND(Command):

    def __init__(self, shx, spline: list) -> None:
        """
        BOND atomnames
        """
        super(BOND, self).__init__(shx, spline)
        _, self.atoms = self._parse_line(spline)


class DISP(Command):
    """
    DISP E f' f"[#] mu[#]
    """

    def __init__(self, shx, spline: list) -> None:
        super(DISP, self).__init__(shx, spline)
        self.element, self.parameter = self._parse_line(spline)


class Restraints():
    """
    Base class for the list of restraints.
    """

    def __init__(self) -> None:
        """
        """
        self._restraints: List[Restraint] = []

    def append(self, restr: Restraint):
        self._restraints.append(restr)

    def __iter__(self) -> Generator:
        x: Restraint
        for x in self._restraints:
            yield x

    def __getitem__(self, item: int) -> Restraint:
        return self._restraints[item]

    def __repr__(self) -> str:
        if self._restraints:
            return "\n".join([str(x) for x in self._restraints])
        else:
            return 'No Restraints in file.'


class DEFS(Restraint):

    def __init__(self, shx, spline: list):
        """
        DEFS sd[0.02] sf[0.1] su[0.01] ss[0.04] maxsof[1]
        Changes the *default* effective standard deviations for the following
        DFIX, SAME, SADI, CHIV, FLAT, DELU and SIMU restraints.
        """
        super(DEFS, self).__init__(shx, spline)
        self.sf = 0.1
        self.su = 0.01
        self.ss = 0.04
        self.maxsof = 1
        self.sd = 0.02
        self.active = True
        p, _ = self._parse_line(spline)
        if _:
            raise ParseParamError(debug=shx.debug, verbose=shx.verbose)
        if len(p) > 0:
            self.sd = p[0]
        if len(p) > 1:
            self.sf = p[1]
        if len(p) > 2:
            self.su = p[2]
        if len(p) > 3:
            self.ss = p[3]
        if len(p) > 4:
            self.maxsof = p[4]

    @property
    def all(self):
        return self.sd, self.sf, self.su, self.ss, self.maxsof


class NCSY(Restraint):
    """
    NCSY DN sd[0.1] su[0.05] atoms
    """

    def __init__(self, shx, spline: list):
        super(NCSY, self).__init__(shx, spline)
        self.sd = 0.1
        self.su = 0.05
        self.DN = None
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.DN = p[0]
        if len(p) > 1:
            self.sd = p[1]
        if len(p) > 2:
            self.su = p[2]
        if not self.DN:
            raise ParseNumError(debug=self.shx.debug, verbose=self.shx.verbose)


class ISOR(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        """
        ISOR s[0.1] st[0.2] atomnames
        """
        super(ISOR, self).__init__(shx, spline)
        self.s = 0.1
        self.st = 0.2
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s = p[0]
        if len(p) > 1:
            self.st = p[1]


class FLAT(Restraint):
    """
    FLAT s[0.1] four or more atoms
    """

    def __init__(self, shx, spline: list):
        super(FLAT, self).__init__(shx, spline)
        self.s = 0.1
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s = p[0]
        # TODO: Have to resolve ranges first:
        # if len(self.atoms) < 4:
        #    raise ParseParamError


class BUMP(Restraint):
    """
    'Anti-bumping' restraints are generated automatically for all distances involving
    two non-bonded C, N, O and S atoms (based on the SFAC type) that are shorter than
    the expected shortest non-bonded distances, allowing for the possibility of hydrogen bonds.
    """

    def __init__(self, shx, spline):
        """
        BUMP s [0.02]
        """
        super(BUMP, self).__init__(shx, spline)
        self.s = 0.02
        p, _ = self._parse_line(spline)
        if len(p) > 0:
            self.s = p[0]
        if _:
            raise ParseParamError(debug=shx.debug, verbose=shx.verbose)


class DFIX(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        DFIX d s[0.02] atom pairs
        """
        super(DFIX, self).__init__(shx, spline)
        self.s = 0.02
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.d = p[0]
        if len(p) > 1:
            self.s = p[1]
        self._paircheck()
        if not self.d:
            raise ParseNumError(debug=self.shx.debug, verbose=self.shx.verbose)
        if (self.shx.debug or self.shx.verbose) and 0.0001 < self.d <= self.s:
            print('*** WRONG ODER of INSTRUCTIONS. d is smaller than s ***')
            print("{}".format(self.textline))


class DANG(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        DANG d s[0.04] atom pairs
        """
        super(DANG, self).__init__(shx, spline)
        self.s = 0.04
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.d = p[0]
        if len(p) > 1:
            self.s = p[1]
        self._paircheck()
        if not self.d:
            raise ParseNumError(debug=self.shx.debug, verbose=self.shx.verbose)
        if 0.0001 < self.d <= self.s:  # Raise exception if d is smaller than s
            raise ParseOrderError(debug=shx.debug, verbose=shx.verbose)


class SADI(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        SADI s[0.02] pairs of atoms
        Instructions with only two atoms are ignored by SHELXL: SADI C3 C4
        SADI_3 C3 C4 C3_4 C4_4 creates SADI C3_3 C4_3  C3_4 C4_4
        SADI_CCF3 C3 C4 C3_4 C4_4 creates SADI C3_1 C4_1  C3_2 C4_2  C3_4 C4_4 if there are residues 1, 2 and 4
        """
        super(SADI, self).__init__(shx, spline)
        self.s = 0.02
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s = p[0]
        self._paircheck()


class SAME(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        SAME s1[0.02] s2[0.04] atomnames
        """
        super(SAME, self).__init__(shx, spline)
        self.s1 = 0.02
        self.s2 = 0.04
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s1 = p[0]
        if len(p) > 1:
            self.s2 = p[1]


class RIGU(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        """
        RIGU s1[0.004] s2[0.004] atomnames
        """
        super(RIGU, self).__init__(shx, spline)
        self.s1 = 0.004
        self.s2 = 0.004
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s1 = p[0]
        if len(p) > 1:
            self.s2 = p[1]


class SIMU(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        """
        SIMU s[0.04] st[0.08] dmax[2.0] atomnames
        """
        super(SIMU, self).__init__(shx, spline)
        self.s = 0.04
        self.st = 0.08
        self.dmax = 2.0
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s = p[0]
        if len(p) > 1:
            self.st = p[1]
        if len(p) > 2:
            self.dmax = p[2]


class DELU(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        """
        DELU s1[0.01] s2[0.01] atomnames
        """
        super(DELU, self).__init__(shx, spline)
        self.s1 = 0.01
        self.s2 = 0.01
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.s1 = p[0]
        if len(p) > 1:
            self.s2 = p[1]


class CHIV(Restraint):

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        """
        CHIV V[0] s[0.1] atomnames
        """
        super(CHIV, self).__init__(shx, spline)
        self.s = 0.1
        self.V = 0.0
        p, self.atoms = self._parse_line(spline)
        if len(p) > 0:
            self.V = p[0]
        if len(p) > 1:
            self.s = p[1]


class EADP(Restraint):
    """
    EADP atomnames
    """

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        super(EADP, self).__init__(shx, spline)
        _, self.atoms = self._parse_line(spline)


class EXYZ(Restraint):
    """
    EADP atomnames
    """

    def __init__(self, shx, spline: list) -> None:
        super(EXYZ, self).__init__(shx, spline)
        _, self.atoms = self._parse_line(spline)


class DAMP(Command):
    """
    DAMP damp[0.7] limse[15]
    """

    def __init__(self, shx, spline: list):
        super(DAMP, self).__init__(shx, spline)
        values, _ = self._parse_line(spline, intnums=False)
        self.damp, self.limse = 0, 0
        if len(values) > 0:
            self.damp = values[0]
        if len(values) > 1:
            self.damp, self.limse = values

    def __repr__(self) -> str:
        if self.limse == 0:
            return "DAMP  {:,g}".format(self.damp)
        else:
            return "DAMP  {:,g} {:,g}".format(self.damp, self.limse)


class HFIX(Command):

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        HFIX mn U[#] d[#] atomnames
        """
        super(HFIX, self).__init__(shx, spline)
        self.params, self.atoms = self._parse_line(spline, intnums=True)

    def __repr__(self):
        return f"HFIX {' '.join([str(x) for x in self.params]) if self.params else ''} " \
               f"{' '.join(self.atoms) if self.atoms else ''}"


class HKLF(Command):

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        HKLF N[0] S[1] r11...r33[1 0 0 0 1 0 0 0 1] sm[1] m[0]
        """
        super(HKLF, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.n = 0
        self.s = 1
        self.matrix = [1, 0, 0, 0, 1, 0, 0, 0, 1]
        self.sm = 1
        self.m = 0
        if len(p) > 0:
            self.n = int(p[0])
        if len(p) > 1:
            self.s = p[1]
        if len(p) > 10:
            self.matrix = p[3:11]
        if len(p) > 11:
            self.sm = p[12]
        if len(p) > 12:
            self.m = p[13]

    def __repr__(self) -> str:
        return "HKLF {:,g} {:,g}  {}  {:,g} {:,g}".format(self.n, self.s, ' '.join([str(i) for i in self.matrix]),
                                                          self.sm, self.m)


class SUMP(Command):
    """
    SUMP for linear equation eypressions with free variables.
    SUMP c sigma c1 m1 c2 m2 ...
    """

    def __init__(self, shx, spline: list):
        super(SUMP, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.c = p.pop(0)
        self.sigma = p.pop(0)
        # this is to have integer free variables
        _fvars: List[int] = [int(x) for x in p[1::2]]
        _times: List[Union[int, float]] = [x for x in p[0::2]]
        self.fvars: List[list[Union[int, float]]] = [[x, y] for x, y in zip(_times, _fvars)]

    def __getitem__(self, item: int) -> Union[List[int], List[float]]:
        return self.fvars[item]


class SWAT(Command):
    """
    SWAT g[0] U[2]
    Allows two variables g and U to be refined in order to model diffuse solvent
    """

    def __init__(self, shx: 'Shelxfile', spline: List[str]):
        super().__init__(shx, spline)
        p, _ = self._parse_line(spline)
        if len(p) > 1:
            self.g = p.pop(0)
            self.U = p.pop(0)


class LATT(Command):
    lattdict = {1: [],  # Primitive
                2: [SymmetryElement(['0.5', '0.5', '0.5'])],  # I-centered
                3: [SymmetryElement(['1/3', '2/3', '2/3']),  # Rhombohedral
                    SymmetryElement(['2/3', '1/3', '1/3'])],
                4: [SymmetryElement(['0.0', '0.5', '0.5']),  # F-centered
                    SymmetryElement(['0.5', '0.0', '0.5']),
                    SymmetryElement(['0.5', '0.5', '0.0'])],
                5: [SymmetryElement(['0.0', '0.5', '0.5'])],  # A-centered
                6: [SymmetryElement(['0.5', '0.0', '0.5'])],  # B-centered
                7: [SymmetryElement(['0.5', '0.5', '0.0'])]}  # C-centered

    lattint_to_str = {1: 'P', 2: 'I', 3: 'R', 4: 'F', 5: 'A', 6: 'B', 7: 'C'}

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        """
        LATT N[1]
        """
        super(LATT, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self.centric = False
        try:
            self.N = int(p[0])
        except ValueError:
            self.N = -1
        self.N_str = self.lattint_to_str[abs(self.N)]
        if self.N > 0:  # centrosymmetric space group:
            self.centric = True
        self.latt_ops = LATT.lattdict[abs(self.N)]


class SYMM(Command):

    def __init__(self, shx, spline: list):
        """
        SYMM symmetry operation
        """
        super(SYMM, self).__init__(shx, spline)
        self.symmcard = self._parse_line(spline)

    def _parse_line(self, spline: list, intnums: bool = False) -> list:
        symmcard = ''.join(spline[1:]).split(',')  # removes whitespace
        return symmcard

    def _as_str(self) -> str:
        return "SYMM  " + ", ".join(self.symmcard)

    def __repr__(self) -> str:
        return self._as_str()

    def __str__(self) -> str:
        return self._as_str()


class SymmCards():
    """
    Contains the list of SYMM cards
    """

    def __init__(self, shx: 'Shelxfile') -> None:
        self.shx = shx
        self._symmcards = [SymmetryElement(['X', 'Y', 'Z'])]
        self.latt_ops = []

    def _as_str(self) -> str:
        return "\n".join([str(x) for x in self._symmcards])

    def __repr__(self) -> str:
        return self._as_str()

    def __str__(self) -> str:
        return self._as_str()

    def __len__(self) -> int:
        return len(self._symmcards)

    def __getitem__(self, item: int) -> SymmetryElement:
        return self._symmcards[item]

    def __iter__(self) -> Generator:
        for x in self._symmcards:
            yield x

    def append(self, symm_data: list) -> None:
        """
        Add the content of a Shelxl SYMM command to generate the appropriate SymmetryElement instance.
        :param symm_data: list of strings. eg.['1/2+X', '1/2+Y', '1/2+Z']
        :return: None
        """
        new_symm = SymmetryElement(symm_data)
        self._symmcards.append(new_symm)
        for symm in self.shx.latt.latt_ops:
            latt_symm = new_symm.apply_latt_symm(symm)
            if latt_symm not in self._symmcards:
                self._symmcards.append(latt_symm)
        if self.shx.latt.centric:
            self._symmcards.append(SymmetryElement(symm_data, centric=True))
            for symm in self.shx.latt.latt_ops:
                latt_symm = new_symm.apply_latt_symm(symm)
                latt_symm.centric = True
                self._symmcards.append(latt_symm)

    def set_centric(self, value: bool) -> None:
        """
        Defines the instance as representing a centrosymmetric structure. Generates the appropriate SymmetryElement
        instances automatically if called before adding further SYMM commands via self.addSymm().
        """
        self.shx.latt.centric = value
        self._symmcards.append(SymmetryElement(['-X', '-Y', '-Z']))
        self._symmcards[-1].centric = True

    def set_latt_ops(self, lattops: list) -> None:
        """
        Adds lattice operations. If called before adding SYMM commands, the appropriate lattice operations are used
        automatically to generate further SymmetryElements.
        :param lattops: list of SymmetryElement instances.
        """
        self.latt_ops = lattops


class LSCycles(Command):
    def __init__(self, shx, spline: list):
        """
        L.S. nls[0] nrf[0] nextra[0]
        #TODO: support nls parameter
        If nrf is positive, it is the number of these cycles that should be performed before applying ANIS.
        Negative nrf indicates which reflections should be ignored during the refinement but used instead for
        the calculation of free R-factors in the final structure factor summation.
        nextra is the number of additional parameters that were derived from the data when 'squeezing' the
        structure etc.
        """
        super(LSCycles, self).__init__(shx, spline)
        p, _ = self._parse_line(spline)
        self._shx = shx
        self.cgls = False
        self._cycles = 0
        self._nrf = ''
        self._nextra = ''
        try:
            self._cycles = int(p[0])
        except (IndexError, NameError, ValueError):
            raise ParseNumError(debug=self.shx.debug, verbose=self.shx.verbose)
        try:
            self._nrf = int(p[1])
        except IndexError:
            pass
        try:
            self._nextra = int(p[2])
        except IndexError:
            pass
        if spline[0].upper() == 'CGLS':
            self.cgls = True

    @property
    def number(self):
        return self._cycles

    @number.setter
    def number(self, n: int) -> None:
        self._cycles = int(n)
        self.__init__(self._shx, self._as_str().split())

    def set_refine_cycles(self, number: int) -> None:
        """
        Sets the number of refinement cycles for the current res file.
        """
        self.number = int(number)

    @property
    def text(self) -> str:
        """
        'CGLS 10 2 '
        """
        return self.__repr__()

    def __iter__(self) -> Generator:
        for x in self.__repr__().split():
            yield x

    def _as_str(self) -> str:
        return '{} {} {} {}'.format('CGLS' if self.cgls else 'L.S.', self._cycles,
                                    self._nrf if self._nrf else '', self._nextra if self._nextra else '').strip()

    def __repr__(self) -> str:
        return self._as_str()


class SFACTable():
    def __init__(self, shx: 'Shelxfile') -> None:
        """
        Holds the information of SFAC instructions. Either with default values and only elements
        SFAC elements
        or as explicit scattering factor in the form of an exponential series, followed by real and
        imaginary dispersion terms, linear absorption coefficient, covalent radius and atomic weight.

        SFAC elements  or  SFAC E a1 b1 a2 b2 a3 b3 a4 b4 c f' f" mu r wt
        """
        self.sfac_table: List[Union[Dict[str, str], Dict[str, int]]] = []
        self.shx = shx
        self.elements_list = []

    def __iter__(self) -> Generator:
        for x in self.sfac_table:
            yield x['element'].capitalize()

    def __repr__(self) -> str:
        sftext = ''
        elements = []
        for sf in self.sfac_table:
            if not self.is_exp(sf) and sf['element'].capitalize() not in elements:
                elements.append(sf['element'].capitalize())
            else:
                if elements:
                    sftext = self._extend_sfac_text(elements, sftext)
                    elements = []
                values = []
                for x in ('element', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'a4', 'b4', 'c',
                          'fprime', 'fdprime', 'mu', 'r', 'wt'):
                    values.append(sf[x])
                sftext = self._extend_sfac_text(elements, sftext)
        if elements:
            sftext = self._extend_sfac_text(elements, sftext)
        return sftext[1:]

    def _extend_sfac_text(self, elements: List[str], sftext: str) -> str:
        sftext += f"\nSFAC {'  '.join(elements)}"
        return sftext

    def __getitem__(self, index: int) -> str:
        """
        Returns the n-th element in the sfac table, beginning with 1.
        """
        if index == 0:
            raise IndexError
        if index < 0:
            index = len(self.sfac_table) + index + 1
        return self.sfac_table[index - 1]['element'].capitalize()

    def parse_element_line(self, spline: List[str]) -> None:
        """
        Adds a new SFAC card to the list of cards.
        """
        if not ''.join(spline[1:]).isalpha():  # joining, because space is not alphabetical
            # Excplicit with all values
            sfdic = {}
            for n, x in enumerate(['element', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'a4', 'b4', 'c',
                                   'fprime', 'fdprime', 'mu', 'r', 'wt']):
                if n == 0:
                    self.elements_list.append(spline[n + 1].upper())
                try:
                    sfdic[x] = spline[n + 1]
                except IndexError:
                    raise ParseNumError(debug=self.shx.debug, verbose=self.shx.verbose)
            self.sfac_table.append(sfdic)
        else:
            # Just the elements
            for x in spline[1:]:
                self.elements_list.append(x.upper())
                self.sfac_table.append({'element': x.upper()})

    def has_element(self, element: str) -> bool:
        return element.upper() in self.elements_list

    @staticmethod
    def is_exp(item: Dict[str, str]) -> bool:
        return 'a1' in item

    def add_element(self, element: str) -> None:
        """
        Adds an element to the SFAC list.
        """
        if self.has_element(element):
            return
        self.elements_list.append(element.upper())
        self.sfac_table.append({'element': element.upper(), 'line_number': None})
        self.shx.unit.add_number(1.0)

    def remove_element(self, element: str):
        del self.sfac_table[self.shx.elem2sfac(element.upper()) - 1]
        del self.elements_list[self.elements_list.index(element.upper())]


class UNIT(Command):
    values: List[Union[int, float]]

    def __init__(self, shx, spline: List):
        """
        UNIT n1 n2 ...
        """
        super(UNIT, self).__init__(shx, spline)
        self.values, _ = self._parse_line(spline)

    def add_number(self, number: float):
        self.values.append(number)

    def __iter__(self):
        yield [x for x in self.values]

    def __repr__(self) -> str:
        return "UNIT " + "  ".join(["{:,g}".format(x) for x in self.values])

    def __str__(self) -> str:
        return self.__repr__()

    def __setitem__(self, key: int, value: Union[int, float]) -> None:
        self.values[key] = value

    def __getitem__(self, item: int) -> Union[int, float]:
        return self.values[item]

    def __add__(self, other: Union[int, float]) -> None:
        self.values.append(other)


class BASF(Command):
    """
    BASF scale factors
    BASF can occour in multiple lines.
    """
    scale_factors: List[Union[int, float]]

    def __init__(self, shx: 'Shelxfile', spline: List[str]) -> None:
        super(BASF, self).__init__(shx, spline)
        self.scale_factors, _ = self._parse_line(spline)

    def __iter__(self):
        yield self.scale_factors


class TWIN(Command):

    def __init__(self, shx, spline: List):
        """
        TWIN 3x3 matrix [-1 0 0 0 -1 0 0 0 -1] N[2]
        +N     -N  m = |N|
        m-1 to 2m-1
        m-1   (2*abs(m)/2)-1
        """
        super(TWIN, self).__init__(shx, spline)
        self.matrix = [-1, 0, 0, 0, -1, 0, 0, 0, -1]
        self.allowed_N = 2
        self.n_value = 2
        if len(spline) > 1:
            p, _ = self._parse_line(spline, intnums=False)
            if len(p) == 9:
                self.matrix = p
            elif len(p) == 10:
                self.matrix = p[:9]
                self.n_value = int(p[9])
            else:
                raise ParseNumError("*** Check TWIN instruction. ***")
        m = abs(self.n_value) / 2
        if self.n_value > 0:
            self.allowed_N = abs(self.n_value) - 1
        else:
            self.allowed_N = (2 * m) - 1


class WGHT(Command):

    def __init__(self, shx, spline: list):
        """
        The weighting scheme is defined as follows:
        w = q / [ σ²(Fo²) + (a*P)² + b*P + d + e*sin(θ)/$lambda; ]

        WGHT a[0.1] b[0] c[0] d[0] e[0] f[.33333]
        Usually only 'WGHT a b' is used
        """
        super(WGHT, self).__init__(shx, spline)
        self.shx = shx
        self.a = 0.1
        self.b = 0.0
        self.c = 0.0
        self.d = 0.0
        self.e = 0.0
        self.f = 0.33333
        p, _ = self._parse_line(spline)
        if len(p) > 0:
            self.a = p[0]
        if len(p) > 1:
            self.b = p[1]
        if len(p) > 2:
            self.c = p[2]
        if len(p) > 3:
            self.d = p[3]
        if len(p) > 4:
            self.e = p[4]
        if len(p) > 5:
            self.f = p[5]

    def _as_string(self):
        wght = 'WGHT   {} {}'.format(self.a, self.b)
        # It is very unlikely that someone changes other parameter than a and b:
        if (self.c + self.d + self.e + self.f) != 0.33333:
            wght += ' {} {} {} {}'.format(self.c, self.d, self.e, self.f)
        return wght

    def difference(self) -> List[float]:
        """
        Returns a list with the weight differences of the parameters a and b.
        """
        try:
            adiff = abs(self.shx.wght.a - self.shx.wght_suggested.a)
            bdiff = abs(self.shx.wght.b - self.shx.wght_suggested.b)
        except AttributeError:
            print("No suggested weighting scheme found. Unable to proceed.")
            return [0.0, 0.0]
        return [round(adiff, 3), round(bdiff, 3)]

    def __repr__(self) -> str:
        return self._as_string()

    def __str__(self) -> str:
        return self._as_string()



from pathlib import Path
from unittest import TestCase

from shelxfile import Shelxfile


class TestShelxFileIsAtom(TestCase):

    def test_is_atom_long_line_true(self):
        """
        """
        is_an_atom = Shelxfile.is_atom(atomline='O1    3    0.120080   0.336659   0.494426  11.00000   0.01445 ...')
        self.assertEqual(True, is_an_atom)

    def test_is_no_atoms_because_type_is_missing(self):
        is_an_atom = Shelxfile.is_atom(atomline='O1    0.120080   0.336659   0.494426  11.00000   0.01445 ...')
        self.assertEqual(False, is_an_atom)

    def test_is_no_atom_because_coordinate_is_missing(self):
        is_an_atom = Shelxfile.is_atom(atomline='O1  4  0.120080    0.494426  11.00000   0.01445 ...')
        self.assertEqual(False, is_an_atom)

    def test_is_no_atom_because_its_a_command(self):
        is_atom = Shelxfile.is_atom("AFIX 123")
        self.assertEqual(False, is_atom)

    def test_is_no_atom_because_its_a_command2(self):
        is_atom = Shelxfile.is_atom("AFIX")
        self.assertEqual(False, is_atom)

    def test_is_an_atom_short(self):
        is_atom = Shelxfile.is_atom('O1    3    0.120080   0.336659   0.494426')
        self.assertEqual(True, is_atom)


class TestShelxfileElementToSfac(TestCase):
    """
    SFAC C H O F Al Ga
    """

    def setUp(self) -> None:
        self.shx = Shelxfile(debug=True)
        self.shx.read_file('tests/resources/p21c.res')

    def test_elem2sfac_oxygen(self):
        self.assertEqual(3, self.shx.elem2sfac('O'))

    def test_elem2sfac_carbon(self):
        self.assertEqual(1, self.shx.elem2sfac('c'))

    def test_elem2sfac_Argon(self):
        self.assertEqual(0, self.shx.elem2sfac('Ar'))


class TestShelxfile(TestCase):

    def setUp(self) -> None:
        self.shx = Shelxfile(debug=True)
        self.shx.read_file('tests/resources/p21c.res')

    def test_sfac2elem_C(self):
        self.assertEqual('C', self.shx.sfac2elem(1))

    def test_sfac2elem_H(self):
        self.assertEqual('H', self.shx.sfac2elem(2))

    def test_sfac2elem_O(self):
        self.assertEqual('O', self.shx.sfac2elem(3))

    def test_sfac2elem_Al(self):
        self.assertEqual('Al', self.shx.sfac2elem(5))

    def test_sfac2elem_not_existent(self):
        self.assertEqual('', self.shx.sfac2elem(8))

    def test_sfac2elem_zero(self):
        self.assertEqual('', self.shx.sfac2elem(0))

    def test_sum_formula(self):
        self.assertEqual('C0.25 H0.5 O0.75 F1 AL1.25 GA1.5', self.shx.sum_formula)

    def test_read_file_to_list(self):
        shx = Shelxfile()
        shx._reslist = ['Foo']
        shx.read_file('tests/resources/p21c.res')
        # Make sure read_file() re-initializes the constructor:
        self.assertEqual(['TITL p21c in P2(1)/c',
                          '    created by SHELXL-2018/3 at 16:18:25 on 03-May-2018',
                          'CELL 0.71073 10.5086 20.9035 20.5072 90 94.13 90',
                          'ZERR 4 0.0003 0.0005 0.0005 0 0.001 0',
                          'LATT 1'],
                         [str(x) for x in shx._reslist[:5]])

    def test_read_string(self):
        self.shx.atoms = 'Foo'
        # Make sure read_string re-initializes the constructor:
        self.shx.read_string(Path('tests/resources/p21c.res').read_text(encoding='latin1'))
        assert self.shx.atoms.number == 148

    def test_read_file(self):
        shx = Shelxfile()
        with self.assertRaises(FileNotFoundError):
            shx.read_file('tests/resources/foobar.res')


class TestShelxfileGoodModel(TestCase):

    def setUp(self) -> None:
        self.shx = Shelxfile()
        self.shx.read_file('tests/resources/model_finished/p21c.res')

    def test_atom_representation(self):
        a = self.shx._reslist[39]
        self.assertEqual(
            'C1    1   -0.187504    0.282782    0.527531    11.00000    0.01807    0.02353      0.01797    0.00008   -0.00179    0.00036',
            str(a))

    def test_atom_object(self):
        a = self.shx._reslist[39]
        self.assertEqual(
            'C1    1   -0.187504    0.282782    0.527531    11.00000    0.01807    0.02353      0.01797    0.00008   -0.00179    0.00036',
            a)

    def test_qpeak(self):
        a = self.shx._reslist[-20]
        self.assertEqual('Q1   1   0.0784    0.1310    0.6428   11.00000  0.04      0.10     ', str(a))

    def test_r1(self):
        assert self.shx.R1 == 0.194

    def test_wr2(self):
        assert self.shx.wr2 == 0.51

    def test_data(self):
        assert self.shx.data == 10786

    def test_param(self):
        assert self.shx.parameters == 551

    def test_data_to_param(self):
        assert round(self.shx.dat_to_param, 4) == 19.5753

    def test_num_restraints(self):
        assert self.shx.num_restraints == 0

    def test_peak_hole(self):
        assert self.shx.highest_peak == 6.395
        assert self.shx.deepest_hole == -5.556

    def test_peak_goof(self):
        assert self.shx.goof == 5.982
        assert self.shx.rgoof == 5.982


class TestWriteFile(TestCase):
    def setUp(self) -> None:
        self.shx = Shelxfile()
        self.shx.read_file('tests/resources/2240189.res')
        self.written = Path('tests/resources/2240189-written.res')
        self.testwritepath = Path('tests/resources/testwrite.res')

    def tearDown(self) -> None:
        self.testwritepath.unlink(missing_ok=True)

    def test_write_shelx_file(self):
        self.shx.write_shelx_file(str(self.testwritepath))
        self.assertEqual(self.written.read_bytes(), self.testwritepath.read_bytes())



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#
import random
import string
from math import sqrt, radians, cos, sin, acos, degrees, floor
from operator import sub, add
from typing import List, Union, Optional, Iterable, Tuple

from shelxfile.misc.misc import flatten, determinante


class Array(object):
    """
    MIT License

    Copyright (c) 2018 Jens Luebben

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.
    """
    __slots__ = ['values']

    def __init__(self, values: Union[list, tuple]):
        self.values = values

    def __iter__(self) -> Iterable[Union[int, float]]:
        for v in self.values:
            yield v

    def __len__(self) -> int:
        return len(self.values)

    def __hash__(self):
        return hash(self.values)

    def __add__(self, other: (list, 'Array')) -> 'Array':
        """
        This method is optimized for speed.
        """
        if isinstance(other, Array):
            return Array(tuple(map(add, self.values, other)))
        elif isinstance(other, (float, int)):
            # return Array( list(map(lambda x: x - other, self.values)) )
            return Array([i + other for i in self.values])
        else:
            raise TypeError(f'Cannot add type Array to type {str(type(other))}.')

    def __iadd__(self, other):
        return self.__add__(other)

    def __eq__(self, other):
        """
        Test for equality.
        """
        return all([a == b for (a, b) in zip(self.values, other.values)])

    def __sub__(self, other):
        """
        Subtracts eiter an Array or a value from the self Array.
        This method is optimized for speed.
        """
        if isinstance(other, Array):
            return Array(list(map(sub, self.values, other)))  # slightly faster
        elif isinstance(other, float) or isinstance(other, int):
            return Array([i - other for i in self.values])
        else:
            raise TypeError('Cannot add type Array to type {}.'.format(str(type(other))))

    def __imul__(self, other):
        """
        Currently supports multiplication by a number.
        __imul__ means a *= b
        """
        if isinstance(other, (int, float)):
            self.values = [v * other for v in self.values]
            return self
        else:
            raise TypeError('Unsupported operation.')

    def __mul__(self, other: ('Array', 'Matrix')) -> (float, 'Array'):
        """
        Calculates: a * b = axbx + ayby + azbz
        """
        if isinstance(other, Matrix):
            # Array() * Matrix()
            a = other[0]
            b = other[1]
            c = other[2]
            x = self.values[0] * a[0] + self.values[1] * b[0] + self.values[2] * c[0]
            y = self.values[0] * a[1] + self.values[1] * b[1] + self.values[2] * c[1]
            z = self.values[0] * a[2] + self.values[1] * b[2] + self.values[2] * c[2]
            return Array([x, y, z])
        else:
            return self.dot(other)

    def __repr__(self):
        return 'Array({})'.format(str(self.values))

    def __getitem__(self, val):
        """
        Get one item from the array.
        """
        return self.values[val]

    def __setitem__(self, pos, val):
        """
        Get one item from the array.
        """
        self.values[pos] = val

    def norm(self):
        """
        The squared lenght of an array
        """
        return sum([n ** 2 for n in self.values])

    def normalized(self):
        """
        Euclidean norm (straight-line distance) of a vector array.
        """
        return sqrt(self.norm())

    @staticmethod
    def zero(m: int) -> 'Array':
        """
        Create zero Array of dimension m
        """
        return Array([0.0 for _ in range(m)])

    @staticmethod
    def randarray(m: int) -> 'Array':
        """
        Create random Array of dimension m
        """
        return Array([random.randint(1, 99) for _ in range(m)])

    @property
    def floor(self):
        return Array(list(map(floor, self.values)))

    def dot(self, other: 'Array') -> float:
        """
        Dot product of an array in kartesian space.
        """
        if len(self) != len(other):
            raise ValueError('Vector sizes must match')
        return sum([i * j for i, j in zip(self, other)])

    def cross(self, other: 'Array') -> 'Array':
        """
        Cross product of the Array (currently only for 3D vectors).
        """
        if len(self) != len(other) != 3:
            raise ValueError('For 3D vectors only')
        a1, a2, a3 = self
        b1, b2, b3 = other
        return Array([(a2 * b3 - a3 * b2), (a3 * b1 - a1 * b3), (a1 * b2 - a2 * b1)])

    def angle(self, other: 'Array') -> float:
        """
        Calculates the angle between two vectors.
        """
        return round(degrees(acos(self.dot(other) / (self.normalized() * other.normalized()))), 9)


class Matrix(object):
    """
    MIT License

    Copyright (c) 2018 Jens Luebben

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    DK: Missing: inverse, eigenvalues
    """
    __slots__ = ['values', 'shape', 'rows', 'columns']

    def __init__(self, values):
        self.shape = (len(values[0]), len(values))
        self.rows = len(values[0])
        self.columns = len(values)
        self.values = values

    def __getitem__(self, val):
        """
        """
        if isinstance(val, (tuple, list)):
            if len(val) == 1:
                return self.values[val[0]]
            return self.values[val[1]][val[0]]
        else:
            return self.values[val]

    def __repr__(self):
        rows = ''
        for row in self.values:
            rows += '|' + ' '.join(['{:>7.4f}'.format(float(x)) for x in row]) + '|' + '\n'
        return rows

    def __add__(self, other: (list, 'Matrix')) -> 'Matrix':
        """
        Matrix addition
        """
        if isinstance(other, Array):
            if len(self) != len(other):
                raise ValueError('Matrix and Array are not of equal length.')
            return Matrix(
                [[sum(e1, e2) for e1, e2 in zip(row1, row2)] for row1, row2 in zip(self.values, other.values)])
        elif isinstance(other, (float, int)):
            return Matrix([[x + other for x in i] for i in self.values])
        elif isinstance(other, Matrix):
            return Matrix([[e1 + e2 for e1, e2 in zip(row1, row2)] for row1, row2 in zip(self.values, other.values)])
        else:
            raise TypeError('Cannot add type {} Array to Matrix.'.format(str(type(other))))

    def __mul__(self, other: ('Matrix', 'Array', int, float)) -> ('Matrix', 'Array'):
        """
        a * b operation
        """
        if isinstance(other, (int, float)):
            return Matrix([[v * other for v in row] for row in self.values])
        elif isinstance(other, (Matrix, OrthogonalMatrix)):
            return Matrix([[sum(ea * eb for ea, eb in zip(a, b)) for b in other.values] for a in self.values])
        elif isinstance(other, Array):
            return Array([sum([b * x for (b, x) in zip(other.values, row)]) for row in self.values])
        else:
            raise TypeError('Cannot add type {} to Matrix.'.format(str(type(other))))

    def __len__(self):
        return self.shape[1]

    def __iter__(self) -> Iterable[List[float]]:
        for n in self.values:
            yield n

    def __eq__(self, other):
        """
        Test for equality.
        """
        return all([b == x for (b, x) in zip(other.values, self.values)])

    def __sub__(self, other):
        """
        Substract two matrices.
        """
        output = []
        for idx in range(len(self)):
            tmp = []
            for val_a, val_b in zip(self[idx], other[idx]):
                tmp.append(val_a - val_b)
            output.append(tmp[:])
        return output[:]

    def __truediv__(self, other: Union['Matrix', float, int]):
        """
        #>>> Matrix([[1, 2, 3], [1, 2, 3], [1, 2, 3]]) / Matrix([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
        A / B = A * B^-1
        """
        if isinstance(other, Matrix):
            return self * other.inversed
        elif isinstance(other, (int, float)):
            return Matrix([[v / other for v in row] for row in self.values])
        else:
            raise NotImplementedError

    def __setitem__(self, key, value):
        # TODO: Implement setitem
        pass

    @property
    def T(self):
        return self.transposed

    @property
    def transposed(self) -> 'Matrix':
        """
        transposes a matrix
        """
        return Matrix(list(zip(*self.values)))

    @property
    def transposed_alt(self) -> 'Matrix':
        """
        Transposes the current matrix.
        """
        rows = []
        for i in range(self.shape[0]):
            rows.append([r[i] for r in self.values])
        return Matrix(rows)

    @property
    def trace(self):
        return self.values[0][0] + self.values[1][1] + self.values[2][2]

    def dot(self, other):
        """
        Dot product of two matrices.
        """
        return Matrix([[sum(a * b for a, b in zip(x_row, y_col)) for y_col in zip(*other)] for x_row in self])

    @staticmethod
    def zero(m: int, n: int) -> 'Matrix':
        """
        Create zero matrix of dimension m,n
        """
        return Matrix([[0.0 for _ in range(n)] for _ in range(m)])

    @staticmethod
    def randmat(m: int, n: int) -> 'Matrix':
        """
        Create random matrix of dimension m, n (rows, columns)

        #>>> Matrix.randmat(5, 3)
        |  2.000   1.000   2.000|
        |  8.000   1.000   2.000|
        |  3.000   5.000   1.000|
        |  4.000   1.000   9.000|
        |  2.000   8.000   4.000|
        <BLANKLINE>
        """
        return Matrix([[random.randint(1, 99) for _ in range(n)] for _ in range(m)])

    def cholesky(self) -> 'Matrix':
        """
        """
        L = Matrix.zero(*self.shape)
        for i, (ai, li) in enumerate(zip(self.values, L.values)):
            for j, lj in enumerate(L.values[:i + 1]):
                s = sum(li[k] * lj[k] for k in range(j))
                li[j] = sqrt(ai[i] - s) if (i == j) else (1.0 / lj[j] * (ai[j] - s))
        return L

    @property
    def inversed(self) -> 'Matrix':
        """
        Inversion of 3 × 3 matrices
        """
        if self.shape != (3, 3):
            raise ValueError('Inversion is only valid for 3x3 Matrix.')
        d = self.det
        m1, m2, m3, m4, m5, m6, m7, m8, m9 = flatten(self.values)
        inv = Matrix([[(m5 * m9 - m6 * m8) / d, (m3 * m8 - m2 * m9) / d, (m2 * m6 - m3 * m5) / d],
                      [(m6 * m7 - m4 * m9) / d, (m1 * m9 - m3 * m7) / d, (m3 * m4 - m1 * m6) / d],
                      [(m4 * m8 - m5 * m7) / d, (m2 * m7 - m1 * m8) / d, (m1 * m5 - m2 * m4) / d]])
        return inv

    @property
    def det(self):
        """
        Return determinant of 3x3 matrix.
        """
        return determinante(self.values)

    @property
    def norm(self):
        return self.frobenius_norm()

    def frobenius_norm(self):
        # To store the sum of squares of the
        # elements of the given matrix
        sumSq = 0
        for i in range(self.rows):
            for j in range(self.columns):
                sumSq += pow(self[i][j], 2)
        # Return the square root of
        # the sum of squares
        return sqrt(sumSq)

    def power_iteration(self, num_simulations=10):
        """
        Eigenvalue algorythm from https://en.wikipedia.org/wiki/Power_iteration

        This does not work. DK
        """
        # Ideally choose a random vector
        # To decrease the chance that our vector
        # Is orthogonal to the eigenvector
        b_k = Matrix.randmat(self.rows, self.columns)
        for _ in range(num_simulations):
            # calculate the matrix-by-vector product Ab
            b_k1 = self.dot(b_k)
            # calculate the norm
            b_k1_norm = b_k1.norm
            # re normalize the vector
            b_k = b_k1 / b_k1_norm
        return b_k


class SymmetryElement(object):
    """
    Class representing a symmetry operation.
    """
    symm_id = 1
    __slots__ = ['centric', 'symms', 'ID', 'matrix', 'trans']

    def __init__(self, symms, centric=False):
        """
        Constructor.
        """
        self.centric = centric
        self.symms = symms
        self.ID = SymmetryElement.symm_id
        SymmetryElement.symm_id += 1
        lines = []
        trans = []
        for symm in self.symms:
            line, t = self._parse_line(symm)
            lines.append(line)
            trans.append(t)
        self.matrix = Matrix(lines).transposed
        self.trans = Array(trans)
        if centric:
            self.matrix *= -1
            self.trans *= -1

    def __str__(self):
        string = "|{aa:2} {ab:2} {ac:2}|   |{v:>4.2}|\n" \
                 "|{ba:2} {bb:2} {bc:2}| + |{vv:>4.2}|\n" \
                 "|{ca:2} {cb:2} {cc:2}|   |{vvv:>4.2}|\n".format(aa=self.matrix[0, 0],
                                                                  ab=self.matrix[0, 1],
                                                                  ac=self.matrix[0, 2],
                                                                  ba=self.matrix[1, 0],
                                                                  bb=self.matrix[1, 1],
                                                                  bc=self.matrix[1, 2],
                                                                  ca=self.matrix[2, 0],
                                                                  cb=self.matrix[2, 1],
                                                                  cc=self.matrix[2, 2],
                                                                  v=float(self.trans[0]),
                                                                  vv=float(self.trans[1]),
                                                                  vvv=float(self.trans[2]))
        return string

    def __repr__(self):
        return self.to_shelxl()

    def __eq__(self, other):
        """
        Check two SymmetryElement instances for equivalence.
        Note that differences in lattice translation are ignored.
        :param other: SymmetryElement instance
        :return: True/False
        """
        m = (self.matrix == other.matrix)
        t1 = Array([v % 1 for v in self.trans])
        t2 = Array([v % 1 for v in other.trans])
        t = (t1 == t2)
        return m and t

    def __sub__(self, other):
        """
        Computes and returns the translational difference between two SymmetryElements. Returns 999.0 if the elements
        cannot be superimposed via an integer shift of the translational parts.
        :param other: SymmetryElement instance
        :return: float
        """
        if self != other:
            return 999.0
        return self.trans - other.trans

    def apply_latt_symm(self, latt_symm):
        """
        Copies SymmetryElement instance and returns the copy after applying the translational part of 'lattSymm'.
        :param latt_symm: SymmetryElement.
        :return: SymmetryElement.
        """
        new_symm = SymmetryElement(self.to_shelxl().split(','))
        new_symm.trans = Array([(self.trans[0] + latt_symm.trans[0]) / 1,
                                (self.trans[1] + latt_symm.trans[1]) / 1,
                                (self.trans[2] + latt_symm.trans[2]) / 1])
        new_symm.centric = self.centric
        return new_symm

    def to_shelxl(self):
        """
        Generate and return string representation of Symmetry Operation in Shelxl syntax.
        :return: string.
        """
        axes = ['X', 'Y', 'Z']
        lines = []
        for i in range(3):
            text = str(self.trans[i]) if self.trans[i] else ''
            for j in range(3):
                s = '' if not self.matrix[i, j] else axes[j]
                if self.matrix[i, j] < 0:
                    s = '-' + s
                elif s:
                    s = '+' + s
                text += s
            lines.append(text)
        return ', '.join(lines)

    def to_cif(self) -> str:
        return self._replace_float_values(self.to_shelxl()).lower()

    def _replace_float_values(self, val: str) -> str:
        val = val.replace('1.25', '5/4')
        val = val.replace('0.75', '3/4')
        val = val.replace('0.5', '1/2')
        val = val.replace('0.33', '1/3')
        val = val.replace('0.25', '1/4')
        val = val.replace('0.125', '1/6')
        return val

    def _parse_line(self, symm: str) -> Tuple[List[int], float]:
        symm = symm.upper().replace(' ', '')
        chars = ['X', 'Y', 'Z']
        line = []
        for char in chars:
            element, symm = self._partition(symm, char)
            line.append(element)
        if symm:
            trans = self._float(symm)
        else:
            trans = 0
        return line, trans

    def _float(self, string: str) -> float:
        try:
            return float(string)
        except ValueError:
            if '/' in string:
                string = string.replace('/', './') + '.'
                return eval('{}'.format(string))

    def _partition(self, symm: str, char: str) -> Tuple[int, str]:
        parts = symm.partition(char)
        if parts[1]:
            if parts[0]:
                sign = parts[0][-1]
            else:
                sign = '+'
            if sign == '-':
                return -1, ''.join((parts[0][:-1], parts[2]))
            else:
                return 1, ''.join((parts[0], parts[2])).replace('+', '')
        else:
            return 0, symm


# End of work by Jens Lübben #############


def my_isnumeric(value: str):
    """
    Determines if a string can be converted to a number.
    """
    try:
        float(value)
    except ValueError:
        return False
    return True


def mean(values):
    """
    returns mean value of a list of numbers
    """
    return sum(values) / float(len(values))


def median(nums):
    """
    calculates the median of a list of numbers
    """
    ls = sorted(nums)
    n = len(ls)
    if n == 0:
        raise ValueError("Need a non-empty iterable")
    # for uneven list length:
    elif n % 2 == 1:
        # // is floordiv:
        return ls[n // 2]
    else:
        return sum(ls[int(int(n) / 2 - 1):int(int(n) / 2 + 1)]) / 2.0


def std_dev(data: List) -> float:
    """
    returns standard deviation of values rounded to pl decimal places
    S = sqrt( (sum(x-xm)^2) / n-1 )
    xm = sum(x)/n
    :param data: list with integer or float values
    """
    if len(data) == 0:
        return 0
    K = data[0]
    n = 0
    summ = 0
    sum_sqr = 0
    for x in data:
        n += 1
        summ += x - K
        sum_sqr += (x - K) * (x - K)
    variance = (sum_sqr - (summ * summ) / n) / (n - 1)
    # use n instead of (n-1) if want to compute the exact variance of the given data
    # use (n-1) if data are samples of a larger population
    return sqrt(variance)


def nalimov_test(data):
    """
    returns a index list of outliers base on the Nalimov test for data.
    Modified implementation of:
    "R. Kaiser, G. Gottschalk, Elementare Tests zur Beurteilung von Messdaten
    Bibliographisches Institut, Mannheim 1972."
    """
    # q-values for degrees of freedom:
    f = {1 : 1.409, 2: 1.645, 3: 1.757, 4: 1.814, 5: 1.848, 6: 1.870, 7: 1.885, 8: 1.895,
         9 : 1.903, 10: 1.910, 11: 1.916, 12: 1.920, 13: 1.923, 14: 1.926, 15: 1.928,
         16: 1.931, 17: 1.933, 18: 1.935, 19: 1.936, 20: 1.937, 30: 1.945}
    fact = sqrt(float(len(data)) / (len(data) - 1))
    fval = len(data) - 2
    if fval < 2:
        return []
    outliers = []
    if fval in f:
        # less strict than the original:
        q_crit = f[fval]
    else:
        q_crit = 1.95
    for num, i in enumerate(data):
        q = abs(((i - median(data)) / std_dev(data)) * fact)
        if q > q_crit:
            outliers.append(num)
    return outliers


def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    """
    returns a random ID like 'L5J74W'
    :param size: length of the string
    :type size: integer
    :param chars: characters used for the ID
    :type chars: string
    """
    return ''.join(random.choice(chars) for _ in range(size))


def atomic_distance(p1: List, p2: List, cell=None, shortest_dist=False):
    """
    p1 and p2 are x, y , z coordinates as list ['x', 'y', 'z']
    cell are the cell parameters as list: ['a', 'b', 'c', 'alpha', 'beta', 'gamma']

    Returns the distance between the two points (Atoms). If shortest_dist is True, the
    shortest distance ignoring translation is computed.
    """
    a, b, c, al, be, ga = 1, 1, 1, 1, 1, 1
    if cell:
        a, b, c = cell[:3]
        al = radians(cell[3])
        be = radians(cell[4])
        ga = radians(cell[5])
    if shortest_dist:
        x1, y1, z1 = [x + 99.5 for x in p1]
        x2, y2, z2 = [x + 99.5 for x in p2]
        dx = (x1 - x2) % 1 - 0.5
        dy = (y1 - y2) % 1 - 0.5
        dz = (z1 - z2) % 1 - 0.5
    else:
        x1, y1, z1 = p1
        x2, y2, z2 = p2
        dx = (x1 - x2)
        dy = (y1 - y2)
        dz = (z1 - z2)
    if cell:
        return sqrt((a * dx) ** 2 + (b * dy) ** 2 + (c * dz) ** 2 + 2 * b * c * cos(al) * dy * dz +
                    2 * dx * dz * a * c * cos(be) + 2 * dx * dy * a * b * cos(ga))
    else:
        return sqrt(dx ** 2 + dy ** 2 + dz ** 2)


def dice_coefficient(a, b, case_insens=True):
    """
    :type a: str
    :type b: str
    :type case_insens: bool
    dice coefficient 2nt/na + nb.
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Dice%27s_coefficient#Python
    """
    if case_insens:
        a = a.lower()
        b = b.lower()
    if not len(a) or not len(b):
        return 0.0
    if len(a) == 1:
        a = a + u'.'
    if len(b) == 1:
        b = b + u'.'
    a_bigram_list = []
    for i in range(len(a) - 1):
        a_bigram_list.append(a[i:i + 2])
    b_bigram_list = []
    for i in range(len(b) - 1):
        b_bigram_list.append(b[i:i + 2])
    a_bigrams = set(a_bigram_list)
    b_bigrams = set(b_bigram_list)
    overlap = len(a_bigrams & b_bigrams)
    dice_coeff = overlap * 2.0 / (len(a_bigrams) + len(b_bigrams))
    return round(dice_coeff, 6)


def dice_coefficient2(a, b, case_insens=True):
    """
    :type a: str
    :type b: str
    :type case_insens: bool
    duplicate bigrams in a word should be counted distinctly
    (per discussion), otherwise 'AA' and 'AAAA' would have a
    dice coefficient of 1...
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Dice%27s_coefficient#Python

    This implementation is reverse. 1 means not hit, 0 means best match
    """
    if case_insens:
        a = a.lower()
        b = b.lower()
    if not len(a) or not len(b):
        return 1.0
    # quick case for true duplicates
    if a == b:
        return 0.0
    # if a != b, and a or b are single chars, then they can't possibly match
    if len(a) == 1 or len(b) == 1:
        return 1.0
    # use python list comprehension, preferred over list.append()
    a_bigram_list = [a[i:i + 2] for i in range(len(a) - 1)]
    b_bigram_list = [b[i:i + 2] for i in range(len(b) - 1)]
    a_bigram_list.sort()
    b_bigram_list.sort()
    # assignments to save function calls
    lena = len(a_bigram_list)
    lenb = len(b_bigram_list)
    # initialize match counters
    matches = i = j = 0
    while i < lena and j < lenb:
        if a_bigram_list[i] == b_bigram_list[j]:
            matches += 2
            i += 1
            j += 1
        elif a_bigram_list[i] < b_bigram_list[j]:
            i += 1
        else:
            j += 1
    score = float(matches) / float(lena + lenb)
    score = 1 - score
    return round(score, 6)


def levenshtein(s1, s2):
    """
    The levensteins distance of two strings.
    """
    s1 = s1.lower()
    s2 = s2.lower()
    if len(s1) < len(s2):
        return levenshtein(s2, s1)
    if len(s2) == 0:
        return len(s1)
    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            # j+1 instead of j since previous_row and current_row are one character longer:
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1  # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    return previous_row[-1]


def distance(x1, y1, z1, x2, y2, z2, round_out=False):
    """
    distance between two points in space for orthogonal axes.
    """
    import math as m
    d = m.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)
    if round_out:
        return round(d, round_out)
    else:
        return d


def vol_unitcell(a, b, c, al, be, ga):
    """
    calculates the volume of a unit cell
    """
    ca, cb, cg = cos(radians(al)), cos(radians(be)), cos(radians(ga))
    v = a * b * c * sqrt(1 + 2 * ca * cb * cg - ca ** 2 - cb ** 2 - cg ** 2)
    return v


class OrthogonalMatrix():
    """
    Orthogonalization matrix used to convert fractional coordinates to cartesian.
    """

    def __init__(self, a, b, c, alpha, beta, gamma):
        self.a, self.b, self.c = a, b, c
        self.V = vol_unitcell(a, b, c, alpha, beta, gamma)
        self.alpha = radians(alpha)
        self.beta = radians(beta)
        self.gamma = radians(gamma)
        self.m = Matrix(((self.a, self.b * cos(self.gamma), self.c * cos(self.beta)),
                         (0, self.b * sin(self.gamma),
                          (self.c * (cos(self.alpha) - cos(self.beta) * cos(self.gamma)) / sin(self.gamma))),
                         (0, 0, self.V / (self.a * self.b * sin(self.gamma)))))
        self.metric_matrix = self.transposed.dot(self.m)
        self._inversed: Optional[Matrix] = None

    def __mul__(self, other: Array) -> Array:
        """
        To convert from fractional to cartesian.
        """
        return self.m * other

    @property
    def inversed(self):
        """
        To convert from cartesian to fractional.
        """
        if not self._inversed:
            self._inversed = self.m.inversed
            return self._inversed
        else:
            return self._inversed

    @property
    def transposed(self):
        return self.m.transposed

    # noinspection PyPep8Naming
    @property
    def T(self):
        return self.m.transposed

    @property
    def values(self):
        return self.m.values


def almost_equal(a: Union[int, float], b: Union[int, float], places=3) -> float:
    """
    Returns True or False if the numbers a and b are equal inside the
    decimal places "places".
    """
    return round(abs(a - b), places) == 0



from unittest import TestCase

from shelxfile import Shelxfile
from shelxfile.misc.dsrmath import Array


class TestOrthogonalMatrix(TestCase):
    def setUp(self) -> None:
        self.shx = Shelxfile(debug=True)
        self.shx.read_file('tests/resources/jkd77.res')

    def test_frac_to_cart(self):
        self.assertListEqual(list(self.shx.atoms.all_atoms[2].cart_coords),
                             list(self.shx.cell.o * Array(self.shx.atoms.all_atoms[2].frac_coords)))

    def test_cart_to_frac(self):
        self.assertListEqual(list(self.shx.atoms.all_atoms[2].frac_coords),
                             [round(x, 8) for x in
                              self.shx.cell.o.m.inversed * Array(self.shx.atoms.all_atoms[2].cart_coords)])

    def test_frac_to_cart_all(self):
        for atom in self.shx.atoms.all_atoms:
            self.assertListEqual(list(atom.cart_coords), list(self.shx.cell.o * Array(atom.frac_coords)))

    def test_cart_to_frac_all(self):
        for atom in self.shx.atoms.all_atoms:
            self.assertListEqual(list(atom.frac_coords),
                                 [round(x, 9) for x in list(self.shx.cell.o.m.inversed * Array(atom.cart_coords))])



class AtomPair():
    def __init__(self, atom1, atom2):
        self.atom1 = atom1
        self.atom2 = atom2

    def __repr__(self):
        if self.atom1 and self.atom2:
            return f'{self.atom1} {self.atom2}'
        else:
            return ''

    def __len__(self):
        if self.atom1 and self.atom2:
            return 2
        elif not self.atom1 or not self.atom2:
            return 0



import datetime
from pathlib import Path
from string import Template
from typing import TYPE_CHECKING, Dict
from shelxfile.version import VERSION
from shelxfile.atoms.atoms import Atoms

if TYPE_CHECKING:
    from shelxfile.shelx.shelxfile import Shelxfile


class CifFile():
    """Class for writing IUCr CIF-1.1 of DDL1 compliant Version 2.4.5 files."""

    def __init__(self, shelx_file: 'Shelxfile', template: str = None):
        if template is not None:
            self.template = template
        else:
            self.template = Path(__file__).parent / "cif_template.tmpl"
        self.data = shelx_file
        self._cif = None

    def __str__(self):
        return self._cif

    def __repr__(self):
        return self._cif

    def write_cif(self, cif_path: Path) -> None:
        with open(self.template, "r") as f:
            template = f.read()
        cif = Template(template)
        sub = cif.substitute(self._cif_dict())
        cif_path.write_text(sub)

    def _cif_dict(self) -> Dict[str, str]:
        cif_dict = {}
        cif_dict["data_name"] = self.data.titl.split()[0].lower() or "unknown"
        cif_dict["version"] = VERSION
        cif_dict["creation_date"] = f"{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        cif_dict["sum_formula"] = self.data.sum_formula
        cif_dict["formula_weight"] = self.data.formula_weight
        cif_dict.update(self._cell_data())
        cif_dict.update(self._symmetry_data())
        cif_dict.update(self._atoms_data())
        cif_dict.update(self._adp_data())
        cif_dict.update(self._misc_dict())
        return cif_dict

    def _cell_data(self) -> Dict[str, float]:
        cell = self.data.cell
        return {
            "cell_a"     : cell.a,
            "cell_b"     : cell.b,
            "cell_c"     : cell.c,
            "cell_alpha" : cell.alpha,
            "cell_beta"  : cell.beta,
            "cell_gamma" : cell.gamma,
            "cell_volume": round(cell.volume, 4),
            "cell_z"     : self.data.zerr.Z,
        }

    def _symmetry_data(self) -> Dict[str, str]:
        symmcards_ = [f" '{sym.to_cif()}'" for sym in self.data.symmcards]
        return {
            "space_group"  : self.data.space_group,
            # "crystal_system": self.data.symmcards.crystal_system,
            "symmetry_loop": '\n'.join(symmcards_),
        }

    def _atoms_data(self) -> Dict[str, str]:
        atoms: Atoms = self.data.atoms
        lines = []
        loop_header = ("loop_\n"
                       " _atom_site_label\n"
                       " _atom_site_type_symbol\n"
                       " _atom_site_fract_x\n"
                       " _atom_site_fract_y\n"
                       " _atom_site_fract_z\n"
                       " _atom_site_adp_type\n"
                       " _atom_site_occupancy\n"
                       " _atom_site_disorder_group")
        for atom in atoms.all_atoms:
            if not atom.qpeak:
                lines.append(f"{atom.fullname_short} {atom.element} "
                             f"{atom.x} {atom.y} {atom.z} "
                             f"{'Uiso' if atom.is_isotropic else 'Uani'} "
                             f"{atom.occupancy} {atom.part.n}")
        return {
            "atom_loop_header": loop_header,
            "atom_loop"       : '\n'.join(lines),
        }

    def _adp_data(self) -> Dict[str, str]:
        atoms: Atoms = self.data.atoms
        lines = []
        loop_header = (
            "loop_\n"
            " _atom_site_aniso_label\n"
            " _atom_site_aniso_U_11\n"
            " _atom_site_aniso_U_22\n"
            " _atom_site_aniso_U_33\n"
            " _atom_site_aniso_U_23\n"
            " _atom_site_aniso_U_13\n"
            " _atom_site_aniso_U_12")
        for atom in atoms.all_atoms:
            if not atom.qpeak and not atom.is_isotropic:
                lines.append(f"{atom.fullname_short} "
                             f"{atom.U11} {atom.U22} {atom.U33} "
                             f"{atom.U23} {atom.U13} {atom.U12} ")
        return {
            "aniso_loop_header": loop_header,
            "aniso_loop"       : '\n'.join(lines),
        }

    def _misc_dict(self) -> Dict[str, str]:
        misc_dict = {}
        misc_dict["temperature"] = round(self.data.temp_in_kelvin, 3) or '?'
        misc_dict["crystal_size_max"] = self.data.size.max or '?'
        misc_dict["crystal_size_mid"] = self.data.size.mid or '?'
        misc_dict["crystal_size_min"] = self.data.size.min or '?'
        misc_dict["wavelength"] = self.data.wavelength or '?'
        misc_dict["R1"] = self.data.R1 or '?'
        misc_dict["wR2"] = self.data.wr2 or '?'
        misc_dict["goodness_of_fit"] = self.data.goof or '?'
        return misc_dict


if __name__ == '__main__':
    from shelxfile import Shelxfile

    shx = Shelxfile(debug=True)
    #shx.read_file('tests/resources/I-43d.res')
    shx.read_file('./tests/resources/p21c.res')

    #cif = CifFile(shx)
    #cif.write_cif(Path('p21c-test.cif'))
    shx.to_cif('p21c-test.cif')



# -*- encoding: utf-8 -*-
# möp
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#
from __future__ import print_function

import os
import re
import subprocess
import sys
from contextlib import suppress
from pathlib import Path
from shutil import which, copyfile

with suppress(ImportError):
    from shelxfile import Shelxfile
from shelxfile.misc.misc import remove_file, sep_line, find_line
from shelxfile.shelx.cards import ACTA


def get_xl_version_string(exe: str) -> str:
    """
    Extracts the version string from a SHELXL executable.
    This is fast and needs no hashes etc.
    :type exe: str
    :param exe: path to SHELXL executable
    """
    try:
        with open(exe, 'rb') as f:
            binary = f.read()
            position = binary.find(b'Version 201')
            if position > 0:
                f.seek(position + 8, 0)  # seek to version string
                version = f.read(6)  # read version string
                return version.decode('ascii')
            else:
                return ''
    except IOError:
        print("Could not determine SHELXL version. Refinement might fail to run.")
        return ''


def find_shelxl_exe(shelxpath=None) -> str:
    """
    returns the appropriate shelxl executable
    """
    names = ['shelxl', 'xl']
    download = 'You can download SHELXL at http://shelx.uni-goettingen.de'
    exe = ''
    if shelxpath:
        return shelxpath
    for name in names:
        exe = which(name)
        if not exe:
            continue
        version = get_xl_version_string(exe)
        if not version:
            print('Your SHELXL version', exe, 'is too old for this Program')
            print('Please use SHELXL 2017 or above!')
            print(download)
        version = version.split('/')
        if int(version[0]) < 2017:
            print('Your SHELXL version is too old. Please use SHELXL 2017 or above!')
            print(download)
        else:
            return exe
    return exe


class ShelxlRefine():
    """
    A class to do a shelxl refinement. It is only for shelxl 2017 and above!
    The resfilename should be without ending.
    """

    def __init__(self, shx: 'Shelxfile', resfile_path: Path, shelxpath: str = None):
        self.shx = shx
        self.shelxpath = shelxpath
        self.resfile_name = resfile_path.stem
        self._shelx_command = find_shelxl_exe(shelxpath)
        self.backup_file = os.path.abspath(str(self.resfile_name + '.shx-bak'))
        self._acta_card = ''  # stores the ACTA values if acta is removed before refinement

        if not self._shelx_command:
            print('\nSHELXL executable not found in system path.\n')
            print('You can download SHELXL at http://shelx.uni-goettingen.de\n')

    def get_b_array(self):
        """
        Approximates the B array size to ensure refinement.
        """
        number_of_atoms = self.shx.atoms.number
        # This is a rough aproximation, but it works until you have a really high number of processors:
        barray = number_of_atoms * 8
        if barray <= 3000:
            barray = 3000
        return barray

    def remove_acta_card(self, acta_card):
        """
        Removes ACTA x from reslist and stores value in self._acta_card.
        """
        if not acta_card:
            return
        self._acta_card = acta_card._textline.strip('\r\n')[:]
        del self.shx._reslist[self.shx.index_of(acta_card)]
        # acta_index = self.shx.index_of(acta_card)
        # self.shx.delete_on_write.update([acta_index])
        self.shx.acta = None

    def restore_acta_card(self):
        """
        Place ACTA after UNIT
        """
        if not self._acta_card:
            return
        acta = ACTA(self.shx, self._acta_card.split())
        self.shx._reslist.insert(self.shx.unit.index + 1, ' ')
        self.shx.acta = self.shx._assign_card(acta, self.shx.unit.index + 1)

    def backup_shx_file(self):
        """
        makes a copy of the res file
        make backup in shxsaves before every fragment fit.
        name: self.resfile_name-date-time-seconds.res
        """
        import datetime
        now = datetime.datetime.now()
        timestamp = (str(now.year) + '_' + str(now.month) + '_' + str(now.day) + '_' +
                     str(now.hour) + '-' + str(now.minute) + '-' + str(now.second))
        resfile = os.path.abspath(str(self.resfile_name + '.res'))
        bakup_dir = os.path.abspath(os.path.dirname(resfile)) + os.path.sep + os.path.relpath('shxsaves')
        try:
            copyfile(resfile, self.backup_file)
        except IOError:
            print('*** Unable to make backup file from {}. ***'.format(resfile))
            sys.exit()
        if not os.path.exists(bakup_dir):
            try:
                os.makedirs(bakup_dir)
            except(IOError, OSError):
                print('*** Unable to create backup directory {}. ***'.format(bakup_dir))
        try:
            copyfile(resfile,
                     bakup_dir + os.path.sep + os.path.split(self.resfile_name)[1] + '_' + timestamp + '.res')
        except IOError:
            print('\n*** Unable to make backup file from {} in shxsaves. ***'.format(resfile))

    def restore_shx_file(self):
        """
        restores filename from backup
        """
        resfile = os.path.abspath(str(self.resfile_name + '.res'))
        try:
            print('*** Restoring previous res file. ***')
            copyfile(self.backup_file, resfile)
        except IOError:
            print('Unable to restore res file from {}.'.format(self.backup_file))
        try:
            remove_file(self.backup_file)
        except IOError:
            print('Unable to delete backup file {}.'.format(self.backup_file))

    def pretty_shx_output(self, out: str):
        """
        selectively prints the output from shelx
        """
        if out.startswith(' +  Copyright(C)'):
            print(' SHELXL {}'.format(' '.join(out.split()[6:8])), end='')
        if out.startswith(' R1'):
            line = out[:].split()
            print(' {}   {} {:>6}'.format(line[0], line[1], line[2][:6]), end='')
        if re.match(r'.*CANNOT RESOLVE (SAME|RIGU|SIMU|DELU)', out):
            print('\nWarning: Are you sure that all atoms are in the correct order?\n')
        if re.match(r'.*CANNOT\s+OPEN\s+FILE.*hkl.*', out):
            print('*** No hkl file found! ***')
            print('*** You need a proper hkl file to run SHELXL! ***')
            sys.exit()
        if re.match(r'.*\*\* Extinction \(EXTI\) or solvent.*', out):
            return
        if re.match(r'.*\*\* MERG code changed to 0', out):
            return
        if re.match(r'.*\*\* Bond\(s\) to .* ignored', out):
            return
        if re.match(r'.*\*\*.*', out):
            print('\n SHELXL says:')
            print(' {}'.format(out.strip('\n\r')), end='')
        if 'before cycle' in out:
            print(out, end='')
        if 'finished at' in out:
            print(out, end='')

    def run_shelxl(self, anis: bool = False, backup_before: bool = True) -> None:
        """
        This method runs shelxl 2013 on the res file self.resfile_name
        """
        if not self._shelx_command:
            print('Unable to refine.')
            return
        if anis:
            self.shx.insert_anis()
        status = True
        resfile = self.resfile_name + '.res'
        hklfile = self.resfile_name + '.hkl'
        current_path = Path('.').resolve()
        # Go into path of resfile:
        os.chdir(Path(resfile).parent)
        if not os.path.exists(hklfile):
            print('You need a proper hkl file to run SHELXL.')
            sys.exit()
        command_line = ['{}'.format(self._shelx_command), "-b{}".format(self.get_b_array()),
                        '{}'.format(self.resfile_name)]
        if backup_before:
            self.backup_shx_file()
        print(sep_line)
        print(' Running SHELXL with "{}" and "{}"'.format(' '.join(command_line), self.shx.cycles))
        with subprocess.Popen(command_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=1,
                              universal_newlines=True) as p:
            for line in p.stdout.readlines():
                # output only the most importand things from shelxl:
                self.pretty_shx_output(line)
        lstfile = Path(f'{self.resfile_name}.lst')
        if lstfile.exists() and lstfile.is_file():
            self.check_refinement_results(lstfile.read_text('latin1').splitlines(keepends=False))
        # Go back to the path before
        os.chdir(current_path)
        if p.returncode != 0:
            status = False
        if os.stat(resfile).st_size < 10:
            # status is False if shelx was unsecessful
            status = False
        if not status:
            print(sep_line)
            print('\nError: SHELXL terminated unexpectedly.')
            print('Check for errors in your SHELX input file!\n')
            self.restore_shx_file()
            sys.exit()

    def check_refinement_results(self, list_file):
        """
        Does some checks if the refinement makes sense e.g. if the data to parameter
        ratio is in an acceptable range.
        """
        regex_final = r' Final Structure Factor Calculation.*'
        final_results = find_line(list_file, regex_final)
        # find data and parameters:
        try:
            dataobj = re.search(r'(\d+\s+data)', list_file[final_results + 4])
            data = float(dataobj.group(0).split()[0])
            parameterobj = re.search(r'(\d+\s+parameters)', list_file[final_results + 4])
            parameters = float(parameterobj.group(0).split()[0])
            restrobj = re.search(r'(\d+\s+restraints)', list_file[find_line(list_file, r" GooF = S =.*")])
            restraints = float(restrobj.group(0).split()[0])
        except AttributeError:
            if self.shx.debug:
                raise
            return False
        try:
            data_to_parameter_ratio = data / parameters
            restr_ratio = ((data + restraints) / parameters)
        except ZeroDivisionError:
            if self.shx.debug:
                raise
            return False
        lattline = find_line(list_file, r'^ LATT.*')
        centro = None
        if lattline:
            try:
                latt = int(list_file[lattline].split()[1])
            except ValueError:
                latt = 1
            if latt > 0:
                centro = True
            else:
                centro = False
        if centro and data_to_parameter_ratio < 10:
            print('*** Warning! The data/parameter ratio is getting low (ratio = {:.1f})! ***'
                  '\n*** but consider (data+restraints)/parameter = {:.1f} ***'
                  .format(data_to_parameter_ratio, restr_ratio))
        if not centro and data_to_parameter_ratio < 7.5:
            print('*** Warning! The data/parameter ratio is getting low (ratio = {:.1f})! ***'
                  '\n*** but consider (data+restraints)/parameter = {:.1f} ***'
                  .format(data_to_parameter_ratio, restr_ratio))


if __name__ == '__main__':
    pass



from contextlib import suppress
from typing import Union, List, Tuple, Optional

from shelxfile.misc import misc

with suppress(Exception):
    from shelxfile import Shelxfile
from shelxfile.misc.dsrmath import atomic_distance, Array, Matrix, SymmetryElement
from shelxfile.misc.elements import get_atomic_number, get_radius_from_element
from shelxfile.misc.misc import split_fvar_and_parameter, ParseSyntaxError, frac_to_cart, ParseUnknownParam
from shelxfile.shelx.cards import PART, AFIX, RESI, CELL, Restraints


class Atom():
    """
    An object holding all Properties of a shelxl atom plus some extra information like
    kartesian coordinates and element type.

    atomname sfac x y z sof[11] U[0.05] or U11 U22 U33 U23 U13 U12
    """
    #                name    sfac     x         y        z       occ      u11      u22 ...
    _anisatomstr = '{:<5s}{:>2}{:>12.6f}{:>12.6f}{:>12.6f}{:>12.5f}{:>11.5f}{:>11.5f}' \
                   ' {:>12.5f}{:>11.5f}{:>11.5f}{:>11.5f}'  # Line wrap is handled during file write.
    #               name    sfac     x         y         z         occ      u11
    _isoatomstr = '{:<5s}{:<2}{:>10.6f}  {:>10.6f}  {:>9.6f}  {:>9.5f}  {:>9.5f}'
    _qpeakstr = '{:<5s}{:<2}{:>8.4f}  {:>8.4f}  {:>8.4f}  {:>9.5f}  {:<9.2f} {:<9.2f}'
    _fragatomstr = '{:<5s}{:>10.6f}  {:>10.6f}  {:>9.6f}'

    def __init__(self, shx: 'Shelxfile') -> None:
        self.shx = shx
        self._cell: CELL = shx.cell
        self.sfac_num: int = 1
        self.resi: Union[RESI, None] = None
        self.part: PART = PART(shx, ['PART', '0'])
        self.afix: Union[AFIX, None] = None
        self._name: str = 'name'  # Name without residue number like "C1"
        # Site occupation factor including free variable like 31.0
        self.sof: float = 11.0
        # fractional coordinates:
        self.x: float = 0.0
        self.y: float = 0.0
        self.z: float = 0.0
        # cartesian coordinates:
        self.xc: float = 0.0
        self.yc: float = 0.0
        self.zc: float = 0.0
        self.qpeak: bool = False
        self.peak_height: float = 0.0
        self.uvals: List[float] = [0.04, 0.0, 0.0, 0.0, 0.0, 0.0]  # [U] or [U11 U22 U33 U23 U13 U12]
        self.U11, self.U22, self.U33, self.U23, self.U13, self.U12 = self.uvals
        self.uvals_orig: List[float] = [0.04, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.frag_atom: bool = False
        self.restraints: List[Restraints] = []
        self._line_numbers = None
        self._occupancy: float = 1.0
        self.molindex: int = 0
        # Indicates if this atom is generated by symmetry:
        self.symmgen: bool = False
        self.pivot: Optional[Atom] = None

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    @property
    def atomid(self) -> int:
        if self.symmgen:
            return 0
        try:
            return self.shx._reslist.index(self)
        except ValueError:
            return 0

    @property
    def fullname(self) -> str:
        # Name including residue number like "C1_2 or C3_0"
        return f'{self.name}_{self.resinum}'

    @property
    def fullname_short(self) -> str:
        if self.resinum == 0:
            return self.name
        else:
            return f'{self.name}_{self.resinum}'

    @property
    def resiclass(self) -> str:
        return self.resi.residue_class

    @property
    def resinum(self) -> int:
        return self.resi.residue_number

    @property
    def chain_id(self) -> int:
        return self.resi.chain_id

    @property
    def fvar(self) -> int:
        # Be aware: fvar can be negative!
        fvar, _ = split_fvar_and_parameter(self.sof)
        return fvar

    @property
    def occupancy(self) -> float:
        # Only the occupancy of the atom like 0.5 (without the free variable)
        _, occ = split_fvar_and_parameter(self.sof)
        # Fractional occupancy:
        if abs(self.fvar) == 1:
            return occ
        else:
            if occ > 0:
                occ = self._get_positive_occupancy(occ)
            else:
                occ = self._get_negative_occupancy(occ)
        return occ

    def _get_negative_occupancy(self, occ):
        try:
            occ = 1 + (self.shx.fvars[self.fvar] * occ)
        except IndexError:
            occ = 1.0
            if self.shx.debug:
                raise ParseSyntaxError(debug=self.shx.debug, verbose=self.shx.verbose)
            if self.shx.verbose:
                print(f'*** Could not get occupancy of free variable {self.fvar} ***')
        return occ

    def _get_positive_occupancy(self, occ):
        try:
            occ = self.shx.fvars[self.fvar] * occ
        except IndexError:
            occ = 1.0  # Happens if the self.fvar is not defined
            if self.shx.debug:
                raise ParseSyntaxError(debug=self.shx.debug, verbose=self.shx.verbose)
            if self.shx.verbose:
                print(f'*** Could not get occupancy of free variable {self.fvar} ***')
        return occ

    @occupancy.setter
    def occupancy(self, occ: float):
        self._occupancy = occ

    @property
    def is_hydrogen(self) -> bool:
        """
        Returns True if the current atom is a hydrogen isotope.
        """
        if self.element in {'H', 'D', 'T'}:
            return True
        else:
            return False

    @property
    def ishydrogen(self) -> bool:
        return self.is_hydrogen

    def set_atom_parameters(self, name: str = 'C', sfac_num: int = 1, coords: List[float] = None, part: PART = None,
                            afix: AFIX = None, resi: RESI = None, site_occupation: float = 11.0,
                            uvals: (list, tuple) = None,
                            symmgen: bool = True):
        """
        Sets atom properties manually if not parsed from a SHELXL file.
        """
        self.name = name
        self.sfac_num = sfac_num
        self.frac_coords = coords
        self.x, self.y, self.z = coords[0], coords[1], coords[2]
        self.xc, self.yc, self.zc = frac_to_cart(self.frac_coords, list(self._cell))
        self.part = part
        self.afix = afix
        self.resi = resi
        self.sof = site_occupation
        self.uvals = uvals
        self.U11, self.U22, self.U33, self.U23, self.U13, self.U12 = uvals
        self.symmgen = symmgen

    def set_uvals(self, uvals: List[float]) -> None:
        """
        Sets u values and checks if a free variable was used.
        """
        self.uvals = uvals
        if sum([abs(x) for x in uvals[2:]]) > 0.000001:  # 0 is Uiso and 1 q-peak height
            # Handle regular atom:
            for n, u in enumerate(uvals):
                if abs(u) > 4.0:
                    fvar, uval = split_fvar_and_parameter(u)
                    # self.uvals[n] = uval
                    self.shx.fvars.set_fvar_usage(fvar)
        else:
            # Handle q-peak case
            if abs(uvals[0]) > 4.0:
                fvar, uval = split_fvar_and_parameter(uvals[0])
                self.shx.fvars.set_fvar_usage(fvar)
        self.U11, self.U22, self.U33, self.U23, self.U13, self.U12 = self.uvals

    def parse_line(self, atline: List, list_of_lines: List, part: PART, afix: AFIX, resi: RESI) -> None:
        """
        Parses the text line of an atom from SHELXL to initialize the atom parameters.
        """
        uvals = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.name = atline[0][:4]  # Atom names are limited to 4 characters
        for n, u in enumerate(atline[6:12]):
            uvals[n] = float(u)
        self.uvals_orig = uvals[:]
        self.set_uvals(uvals)
        self._line_numbers = list_of_lines
        self.part = part
        self.afix = afix
        self.resi = resi
        self._get_part_and_occupation(atline)
        self.x, self.y, self.z = self._get_atom_coordinates(atline)
        self.xc, self.yc, self.zc = self._cell.o * Array(self.frac_coords)
        if abs(self.uvals[1]) > 0.0 and abs(
                self.uvals[2]) < 0.000001 and self.shx.hklf:  # qpeaks are always behind hklf
            self.peak_height = uvals[1]
            self.qpeak = True
        if self.shx.end:  # After 'END' can only be Q-peaks!
            self.qpeak = True
        self.sfac_num = int(atline[1])
        self.shx.fvars.set_fvar_usage(self.fvar)
        # TODO: I am still unsure if this these are correct values:
        # self.Uiso = self.Ueq
        # transformed_u = self.transform_u_by_symmetry(2)
        # print(self.name, [round(x, 6) for x in transformed_u], self.frac_coords)

    @property
    def ucif(self):
        return self.set_ucif(self.uvals)

    @property
    def ustar(self):
        return self.ucif * self._cell.N * self._cell.N.T

    @property
    def u_cart(self):
        return self.ustar * self._cell.o * self._cell.o.T

    @property
    def ueq(self):
        return self.set_ueq(self.uvals)

    def set_ueq(self, uvals):
        # This is a q-peak:
        if uvals[0] > 0 and not sum(uvals[2:]):
            ueq = uvals[0]
        # This is a hydrogen atom with negative thermal parameter:
        # elif uvals[0] < 0 and not sum(uvals[1:]):
        #    ueq = self.pivot.Uiso * abs(uvals[0])
        else:
            # This is a non-hydrogen atom with an ADP
            ueq = self.u_cart.trace / 3
        return ueq

    def set_ucif(self, uvals):
        """
        SHELXL uses U(cif)
        U(star) = N * U(cif) * N.T
        U(cart) = A * U(star) * A.T
        U(star) = R * U(star) * R^t
        U(cif) = N^-1 * U(star) * (N^-1).T
        U(star) = A^-1 * U(cart) * A^-1.T
        """
        U11, U22, U33, U23, U13, U12 = uvals
        U21 = U12
        U32 = U23
        U31 = U13
        return Matrix([[U11, U12, U13], [U21, U22, U23], [U31, U32, U33]])

    def transform_u_by_symmetry(self, symmetry_number: int):
        symm = '-y, +x-y, +z'.split(',')
        R = SymmetryElement(symm).matrix
        Ustar_n = self.ustar * R * R.T
        Ucif_n = Ustar_n * self._cell.N.inversed * self._cell.N.inversed.T
        uvals = Ucif_n
        upper_diagonal = uvals.values[0][0], uvals.values[1][1], uvals.values[2][2], \
            uvals.values[1][2], uvals.values[0][2], \
            uvals.values[0][1]
        return upper_diagonal

    def _get_part_and_occupation(self, atline: List[str]) -> None:
        # TODO: test all variants of PART and AFIX sof combinations:
        if self.part.sof != 11.0:
            if self.afix and self.afix.sof:  # handles position of afix and part:
                if self.afix.index > self.part.index:
                    self.sof = self.afix.sof
            else:
                self.sof = self.part.sof
        elif self.afix and self.afix.sof:
            if self.part.sof != 11.0:
                if self.part.index > self.afix.index:
                    self.sof = self.part.sof
            else:
                self.sof = self.afix.sof
        elif len(atline) > 5:
            self.sof = float(atline[5])

    def _get_atom_coordinates(self, atline: List[str]) -> Tuple[float, float, float]:
        try:
            x, y, z = [float(x) for x in atline[2:5]]
        except ValueError as e:
            if self.shx.debug or self.shx.verbose:
                print(e, 'Line:', self._line_numbers[-1])
                raise ParseUnknownParam(debug=self.shx.debug, verbose=self.shx.verbose)
            else:
                x, y, z = 0.1, 0.1, 0.1
        if abs(x) > 4:
            fvar, x = split_fvar_and_parameter(x)
            self.shx.fvars.set_fvar_usage(fvar)
        if abs(y) > 4:
            fvar, y = split_fvar_and_parameter(y)
            self.shx.fvars.set_fvar_usage(fvar)
        if abs(z) > 4:
            fvar, z = split_fvar_and_parameter(z)
            self.shx.fvars.set_fvar_usage(fvar)
        return x, y, z

    @property
    def is_isotropic(self) -> bool:
        return sum(self.uvals[1:]) == 0

    @property
    def uvals_as_list(self) -> List[list[float]]:
        U11, U22, U33, U23, U13, U12 = self.uvals
        U21 = U12
        U32 = U23
        U31 = U13
        return [[U11, U12, U13], [U21, U22, U23], [U31, U32, U33]]

    def is_npd(self) -> bool:
        eigenvalues = misc.eigenvals(self.u_cart.values)
        if any(ev <= 0 for ev in eigenvalues):
            return True
        else:
            return False

    @property
    def element(self) -> str:
        """
        Chemical element character
        """
        return self.shx.sfac2elem(self.sfac_num).capitalize()

    @element.setter
    def element(self, new_element: str) -> None:
        """
        Sets the element type of atom.
        """
        sfac = self.shx.elem2sfac(new_element)
        if sfac == 0:
            self.shx.sfac_table.add_element(new_element)
            sfac = self.shx.elem2sfac(new_element)
        self.sfac_num = sfac

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        name_has_residue_number = new_name.split('_')[-1].isdigit()
        if '_' in new_name and name_has_residue_number:
            print('*** Illegal atom name. Use a name without a residue number. ***')
            return
        self._name = new_name

    @property
    def an(self) -> int:
        return get_atomic_number(self.element)

    @property
    def radius(self) -> float:
        """
        Returns the atomic covalence radius in angstrom.
        """
        return get_radius_from_element(self.element)

    def __iter__(self):
        for x in self.__repr__().split():
            yield x

    def __repr__(self) -> str:
        return 'Atom ID: {}'.format(self.atomid)

    def __str__(self) -> str:
        """
        Returns a text line of the Atom with SHELXL syntax.
        :return: SHELX-formated atom string
        """
        if self.afix and self.shx.frag:
            # An atom from a FRAG/FEND instruction
            return Atom._fragatomstr.format(self.name, self.x, self.y, self.z)
        else:
            if sum([abs(u) for u in self.uvals[2:]]) > 0.00001 and not self.qpeak:
                # anisotropic atom
                try:
                    return Atom._anisatomstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof,
                                                    *self.uvals)
                except IndexError:
                    return 'REM Error in U values.'
            else:
                # isotropic atom:
                if self.qpeak:
                    return Atom._qpeakstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof, 0.04,
                                                 self.peak_height)
                try:
                    return Atom._isoatomstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof,
                                                   *self.uvals)
                except IndexError:
                    return Atom._isoatomstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof, 0.04)

    @property
    def index(self) -> int:
        # The position in the res file as index number (starting from 0).
        return self.shx.index_of(self)

    @property
    def frac_coords(self) -> tuple:
        return self.x, self.y, self.z

    @frac_coords.setter
    def frac_coords(self, coords: List[float]):
        self.x, self.y, self.z = coords

    @property
    def cart_coords(self) -> Tuple[float, float, float]:
        return self.xc, self.yc, self.zc

    def delete(self):
        """
        Delete atom(s) in the file.
        """
        del self.shx.atoms[self.index]
        self.shx.atoms._atomsdict.clear()

    def to_isotropic(self) -> None:
        """
        Makes the current atom isotropic.
        """
        self.uvals = [0.04, 0.0, 0.0, 0.0, 0.0, 0.0]

    def find_atoms_around(self, dist=1.2, only_part=0) -> List['Atom']:
        """
        Finds atoms around the current atom.
        """
        found = []
        for at in self.shx.atoms:
            if atomic_distance([self.x, self.y, self.z], [at.x, at.y, at.z], self._cell) < dist \
                    and self != at and at.part.n == only_part and not at.qpeak:
                # only in special part and no q-peaks:
                found.append(at)
        return found

    r'''
    def get_pivot_atom(self) -> Union['Atom', None]:
        """
        Returns the pivot atom (C1) of a riding hydrogen atom e.g. (H1, H2, or H3).
           /H1
        -C1--H2
           \H3
        """
        pivots = self.find_atoms_around(dist=1.2)
        return pivots[0] if pivots and self.afix.mn else None'''
